# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the GRO reader.
"""
# Pylint complains about hypothesis strategies not receiving a value for the
# `draw` parameter. This is because the `draw` parameter is implicitly filled
# by hypothesis. The pylint warning is disabled for the file instead of being
# disabled at every call of a strategy.
# pylint: disable=no-value-for-parameter

from pprint import pprint
import copy
import collections
import itertools

import numpy as np
import networkx as nx

import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp

import vermouth
from vermouth.file_writer import DeferredFileWriter
from vermouth.utils import are_different
from vermouth.molecule import Molecule
from vermouth.gmx import gro

# pylint: disable=redefined-outer-name


# The data comes from residues 3 to 5 of 1BTA.pdb. The atoms from ILE 5 are
# changed in solvent.
GRO_CONTENT = """\
    1ALA      N    1
    1ALA     CA    2
    1ALA      C    3
    1ALA      O    4
    1ALA     CB    5
    1ALA      H    6
    1ALA     HA    7
    1ALA    HB1    8
    1ALA    HB2    9
    1ALA    HB3   10
    2VAL      N   11
    2VAL     CA   12
    2VAL      C   13
    2VAL      O   14
    2VAL     CB   15
    2VAL    CG1   16
    2VAL    CG2   17
    2VAL      H   18
    2VAL     HA   19
    2VAL     HB   20
    2VAL   HG11   21
    2VAL   HG12   22
    2VAL   HG13   23
    2VAL   HG21   24
    2VAL   HG22   25
    2VAL   HG23   26
    3SOL     OW   27
    3SOL    HW1   28
    3SOL    HW2   29
    4SOL     OW   30
    4SOL    HW1   31
    4SOL    HW2   32
    5SOL     OW   33
    5SOL    HW1   34
    5SOL    HW2   35
    6SOL     OW   36
    6SOL    HW1   37
    6SOL    HW2   38
    7SOL     OW   39
    7SOL    HW1   40
    7SOL    HW2   41
    8SOL     OW   42
    8SOL    HW1   43
    8SOL    HW2   44
""".split('\n')

_COORDINATES_STR = """\
  -0.204   0.447   0.725
  -0.109   0.340   0.685
   0.032   0.401   0.675
   0.047   0.521   0.659
  -0.149   0.283   0.549
  -0.199   0.535   0.683
  -0.109   0.262   0.759
  -0.226   0.346   0.505
  -0.063   0.281   0.483
  -0.187   0.183   0.561
   0.134   0.321   0.686
   0.273   0.377   0.678
   0.364   0.277   0.607
   0.384   0.167   0.654
   0.325   0.402   0.819
   0.462   0.471   0.812
   0.226   0.491   0.895
   0.121   0.226   0.700
   0.271   0.470   0.623
   0.335   0.308   0.871
   0.455   0.557   0.747
   0.491   0.503   0.911
   0.535   0.401   0.773
   0.178   0.559   0.825
   0.151   0.430   0.942
   0.279   0.549   0.969
   0.421   0.315   0.496
   0.512   0.221   0.424
   0.657   0.257   0.455
   0.718   0.339   0.389
   0.488   0.230   0.274
   0.336   0.223   0.248
   0.562   0.114   0.206
   0.303   0.125   0.135
   0.404   0.405   0.460
   0.492   0.120   0.458
   0.527   0.324   0.237
   0.286   0.191   0.338
   0.301   0.321   0.221
   0.658   0.101   0.253
   0.504   0.024   0.216
   0.576   0.137   0.101
   0.378   0.133   0.057
   0.304   0.024   0.174\
"""
COORDINATES = [[float(x) for x in line.split()]
               for line in _COORDINATES_STR.split('\n')]
VELOCITIES = [[x + 1 for x in line] for line in COORDINATES]

DIFFERENCE_USE_CASE = (
    '', [], 'a', 2, 3, True, False, None, float('nan'), float('inf'),
    [True], [False], [True, False], [False, False], [None], [None, None],
    [1.1, 1.1], [0.0, 1.1],
    [float('nan'), float('inf')],
    [float('inf'), float('nan')],
    ['a', 'b'], 'ab',
    np.array([1.2, 2.3, 4.5]),
    np.array([[1.2, 2.3, 4.5], [5.6, 7.8, 8.9]]),
    np.array([np.nan, np.inf, 3]),
    [[6, 8], [9, 10, 11]],
    {'a': 3, 2: 'tata'},
    {'a': 3, 2: np.array([2, 3, 4])},
    {'a': 4, 2: 'not tata'},
)


@pytest.fixture(params=[True, False])
def gro_reference(request, tmpdir_factory):
    """
    Generate a GRO file and the corresponding molecule.
    """
    filename = tmpdir_factory.mktemp("data").join("tmp.gro")
    with open(str(filename), 'w') as outfile:
        write_ref_gro(outfile, velocities=request.param, box='10.0 11.1 12.2')
    molecule = build_ref_molecule(velocities=request.param)
    return filename, molecule


@pytest.fixture(params=[43, 45])
def gro_wrong_length(request, gro_reference, tmpdir_factory):  # pylint: disable=redefined-outer-name
    """
    Generate a GRO file with a wrong number of atoms on line 2.
    """
    path_in, _ = gro_reference
    path_out = tmpdir_factory.mktemp("data").join("wrong.gro")
    with open(str(path_in)) as infile, open(str(path_out), 'w') as outfile:
        outfile.write(next(infile))
        outfile.write('{}\n'.format(request.param))
        for line in infile:
            outfile.write(line)
    return path_out


def write_ref_gro(outfile, velocities=False, box='10.0 10.0 10.0'):
    """
    Write a GRO file from the reference data.

    Parameters
    ----------
    outfile: file
        The file in which to write.
    velocities: bool
        Set if velocities must be written.
    box: str
        The box string to write at the end of the file.
    """
    velocity_fmt = ''
    if velocities:
        velocity_fmt = '{:8.4f}' * 3
    outfile.write('Just a title\n')
    outfile.write(str(len(COORDINATES)) + '\n')
    for atom, coords, vels in zip(GRO_CONTENT, COORDINATES, VELOCITIES):
        outfile.write(('{}{:8.3f}{:8.3f}{:8.3f}' + velocity_fmt + '\n')
                      .format(atom, *itertools.chain(coords, vels)))
    outfile.write(box)
    outfile.write('\n')


def build_ref_molecule(velocities=False):
    """
    Build the molecule graph corresponding to the reference data.

    Parameters
    ----------
    velocities: bool
        Set if velocities must be written.

    Return
    ------
    Molecule
    """
    nodes = (
        {'resid': 1, 'resname': 'ALA', 'atomname': 'N', 'atomid': 1, 'element': 'N'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'CA', 'atomid': 2, 'element': 'C'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'C', 'atomid': 3, 'element': 'C'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'O', 'atomid': 4, 'element': 'O'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'CB', 'atomid': 5, 'element': 'C'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'H', 'atomid': 6, 'element': 'H'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'HA', 'atomid': 7, 'element': 'H'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'HB1', 'atomid': 8, 'element': 'H'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'HB2', 'atomid': 9, 'element': 'H'},
        {'resid': 1, 'resname': 'ALA', 'atomname': 'HB3', 'atomid': 10, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'N', 'atomid': 11, 'element': 'N'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'CA', 'atomid': 12, 'element': 'C'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'C', 'atomid': 13, 'element': 'C'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'O', 'atomid': 14, 'element': 'O'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'CB', 'atomid': 15, 'element': 'C'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'CG1', 'atomid': 16, 'element': 'C'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'CG2', 'atomid': 17, 'element': 'C'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'H', 'atomid': 18, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HA', 'atomid': 19, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HB', 'atomid': 20, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG11', 'atomid': 21, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG12', 'atomid': 22, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG13', 'atomid': 23, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG21', 'atomid': 24, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG22', 'atomid': 25, 'element': 'H'},
        {'resid': 2, 'resname': 'VAL', 'atomname': 'HG23', 'atomid': 26, 'element': 'H'},
        {'resid': 3, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 27, 'element': 'O'},
        {'resid': 3, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 28, 'element': 'H'},
        {'resid': 3, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 29, 'element': 'H'},
        {'resid': 4, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 30, 'element': 'O'},
        {'resid': 4, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 31, 'element': 'H'},
        {'resid': 4, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 32, 'element': 'H'},
        {'resid': 5, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 33, 'element': 'O'},
        {'resid': 5, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 34, 'element': 'H'},
        {'resid': 5, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 35, 'element': 'H'},
        {'resid': 6, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 36, 'element': 'O'},
        {'resid': 6, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 37, 'element': 'H'},
        {'resid': 6, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 38, 'element': 'H'},
        {'resid': 7, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 39, 'element': 'O'},
        {'resid': 7, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 40, 'element': 'H'},
        {'resid': 7, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 41, 'element': 'H'},
        {'resid': 8, 'resname': 'SOL', 'atomname': 'OW', 'atomid': 42, 'element': 'O'},
        {'resid': 8, 'resname': 'SOL', 'atomname': 'HW1', 'atomid': 43, 'element': 'H'},
        {'resid': 8, 'resname': 'SOL', 'atomname': 'HW2', 'atomid': 44, 'element': 'H'},
    )
    for node, coords, vels in zip(nodes, COORDINATES, VELOCITIES):
        node['chain'] = ''
        node['position'] = np.array(coords)
        if velocities:
            node['velocity'] = np.array(vels)
    molecule = Molecule()
    molecule.add_nodes_from(enumerate(nodes))
    return molecule


def filter_molecule(molecule, exclude, ignh):
    """
    Remove nodes from a graph based on some criteria.

    Nodes are removed if their resname matches one from
    the `exclude` argument, or of `ighn` is `True` and the node is a
    hydrogen.

    The atoms are renumbered after the filtered atoms are removed. This will
    scramble the edges and the interactions! This function is meant to be used
    on molecules with ONLY nodes.

    Parameters
    ----------
    molecule: nx.Graph
        The molecule to filter. The molecule is modified in place.
    exclude: list
        List of residue name to exclude.
    ighn: bool
        If `True`, the hydrogens are excluded.
    """
    to_remove = []
    exclusions = [('resname', value) for value in exclude]
    if ignh:
        exclusions.append(('element', 'H'))
    for node_key, node in molecule.nodes.items():
        for key, value in exclusions:
            if key in node and node[key] == value:
                to_remove.append(node_key)
                break
    molecule.remove_nodes_from(to_remove)
    mapping = collections.OrderedDict(
        [(old, new) for new, old in enumerate(molecule.nodes)]
    )
    nx.relabel_nodes(molecule, mapping, copy=False)
    # Networkx mangles the order of the nodes, but we want the nodes sorted.
    new_nodes = sorted(molecule.nodes.items())
    molecule.clear()  # Remove all the nodes
    molecule.add_nodes_from(new_nodes)


def compare_dicts(testee, reference):
    """
    Report differences between dictionaries.
    """
    report = []
    dict_equals = testee.keys() == reference.keys()
    diff_keys = []
    if dict_equals:
        for key in testee:
            if are_different(testee[key], reference[key]):
                diff_keys.append(key)
    if diff_keys:
        report.append('The following keys differ: {}'.format(diff_keys))
    if diff_keys or not dict_equals:
        report.append('+ {}'.format(testee))
        report.append('- {}'.format(reference))
    return report


def assert_molecule_equal(molecule, reference):
    """
    Make a pytest test fail with a report if the molecules are not equal.
    """
    report = []
    # Nodes
    if len(molecule.nodes) != len(reference.nodes):
        report.append(
            'Different number of nodes: {} != {}'
            .format(len(molecule.nodes), len(reference.nodes))
        )
    zipped_nodes = zip(molecule.nodes.items(), reference.nodes.items())
    for (key_mol, node_mol), (key_ref, node_ref) in zipped_nodes:
        partial_report = []
        failed = False
        if key_mol != key_ref:
            partial_report.append('Node key {} != {}'.format(key_mol, key_ref))
            failed = True
        partial_report = compare_dicts(node_mol, node_ref)
        if partial_report:
            if not failed:
                report.append('Node {}'.format(key_mol))
            report.extend(partial_report)
        elif failed:
            report.append('= {}'.format(node_mol))

    extra_molecule_nodes = molecule.nodes.keys() - reference.nodes.keys()
    if extra_molecule_nodes:
        report.append('The following nodes are not in the reference: {}.'
                      .format(extra_molecule_nodes))
    extra_reference_nodes = reference.nodes.keys() - molecule.nodes.keys()
    if extra_reference_nodes:
        report.append('The following nodes are not in the molecule: {}.'
                      .format(extra_reference_nodes))
    # Edges
    # The order of the edges does not matter.
    molecule_edges = set(molecule.edges)
    reference_edges = set(reference.edges)
    extra_molecule_edges = molecule_edges - reference_edges
    extra_reference_edges = reference_edges - molecule_edges
    if extra_molecule_edges:
        report.append('The following edges are not in the reference:')
        report.append('{}'.format(extra_molecule_edges))
    if extra_reference_edges:
        report.append('The following edges are not in the molecule:')
        report.append('{}'.format(extra_reference_edges))
    # Meta data
    report.extend(compare_dicts(molecule.meta, reference.meta))
    if molecule._force_field is not reference._force_field:  # pylint: disable=protected-access
        report.append('Force fields do not match: {} != {}'
                      .format(molecule._force_field, reference._force_field))  # pylint: disable=protected-access
    if molecule.nrexcl != reference.nrexcl:
        report.append('nrexcl: {} != {}'.format(molecule.nrexcl, reference.nrexcl))
    # Interactions
    if report:
        pytest.fail('\n'.join(report))


@st.composite
def generate_dict(draw, min_size=0):
    """
    Strategy to generate an arbitrary dictionary.
    """
    keys = st.one_of(st.text(), st.integers())
    values = st.one_of(
        st.text(), st.integers(), st.floats(),
        hnp.arrays(dtype=hnp.scalar_dtypes(), shape=hnp.array_shapes())
    )
    dict_a = draw(st.dictionaries(keys, values, min_size=min_size, max_size=10))
    return dict_a


@st.composite
def generate_equal_dict(draw):
    """
    Strategy to generate two equal dictionaries.
    """
    dict_a = draw(generate_dict())
    dict_b = copy.deepcopy(dict_a)
    return dict_a, dict_b



@st.composite
def generate_diff_dict(draw):
    """
    Strategy to generate to similar but different dictionaries.
    """
    dict_a = draw(generate_dict(min_size=1))
    dict_b = copy.deepcopy(dict_a)
    values = st.one_of(
        st.text(), st.integers(), st.floats(),
        hnp.arrays(dtype=hnp.scalar_dtypes(), shape=hnp.array_shapes())
    )
    num_to_change = draw(st.integers(
        min_value=1, max_value=(len(dict_b) // 2) + 1
    ))
    for _ in range(num_to_change):
        key = draw(st.sampled_from(list(dict_b.keys())))
        new_val = draw(values)
        assume(are_different(new_val, dict_a[key]))
        dict_b[key] = new_val
    dict_c = draw(generate_dict())
    for key in set(dict_c.keys()) - set(dict_b.keys()):
        dict_b[key] = dict_c[key]
    return dict_a, dict_b


@given(generate_equal_dict())
def test_compare_dict_equal(dict_a_and_b):
    """
    Test that :func:`compare_dicts` identify equal dictionaries.
    """
    dict_a = dict_a_and_b[0]
    dict_b = dict_a_and_b[1]
    assert not compare_dicts(dict_a, dict_b)  # report is empty


@given(generate_diff_dict())
def test_compare_dict_diff(dict_a_and_b):
    """
    Test that :func:`compare_dicts` identify different dictionaries.
    """
    dict_a = dict_a_and_b[0]
    dict_b = dict_a_and_b[1]
    assert compare_dicts(dict_a, dict_b)  # report is not empty


def test_filter_molecule_none(gro_reference):  # pylint: disable=redefined-outer-name
    """
    Test that :func:`filter_molecule` does not remove nodes when not asked for.
    """
    _, molecule = gro_reference
    filter_molecule(molecule, exclude=(), ignh=False)
    assert len(molecule.nodes) == 44


def test_filter_molecule_ignh(gro_reference):  # pylint: disable=redefined-outer-name
    """
    Test that :func:`filter_molecule` works with `ighn` argument.
    """
    _, molecule = gro_reference
    filter_molecule(molecule, exclude=(), ignh=True)
    elements = [node['element'] for node in molecule.nodes.values()]
    assert len(molecule.nodes) == (44 - 26)
    assert 'H' not in elements


@pytest.mark.parametrize('exclude', (
    ('SOL', ), ('ALA', ), ('ALA', 'VAL'),
    ('ALA', 'VAL', 'SOL'), ('XXX', ), ('ALA', 'XXX'),
))
def test_filter_molecule_exclude(gro_reference, exclude):  # pylint: disable=redefined-outer-name
    """
    Test that :func:`filter_molecule` works with `exclude` argument.
    """
    _, molecule = gro_reference
    filter_molecule(molecule, exclude=exclude, ignh=False)
    resnames = [node['resname'] for node in molecule.nodes.values()]
    should_not_be_there = [exclusion
                           for exclusion in exclude
                           if exclusion in resnames]
    assert not should_not_be_there


def test_filter_molecule_identity(gro_reference):  # pylint: disable=redefined-outer-name
    """
    Test that :func:`filter_molecule` modifies the molecule in place.
    """
    _, molecule = gro_reference
    id_before = id(molecule)
    filter_molecule(molecule, exclude=('SOL', ), ignh=True)
    assert id(molecule) == id_before


def test_filter_molecule_order(gro_reference):  # pylint: disable=redefined-outer-name
    """
    Test that :func:`filter_molecule` returns the nodes in order.
    """
    _, molecule = gro_reference
    filter_molecule(molecule, exclude=('SOL', ), ignh=True)
    keys = list(molecule.nodes)
    assert  sorted(keys) == keys


@pytest.mark.parametrize('exclude', (
    (), ('SOL', ), ('ALA', ), ('ALA', 'VAL'),
    ('ALA', 'VAL', 'SOL'), ('XXX', ), ('ALA', 'XXX'),
))
@pytest.mark.parametrize('ignh', (True, False))
def test_read_gro(gro_reference, exclude, ignh):  # pylint: disable=redefined-outer-name
    """
    Test the GRO reader.
    """
    filename, reference = gro_reference
    filter_molecule(reference, exclude=exclude, ignh=ignh)
    molecule = gro.read_gro(filename, exclude=exclude, ignh=ignh)
    pprint(list(molecule.nodes.items()))
    assert_molecule_equal(molecule, reference)


def test_read_gro_wrong_atom_number(gro_wrong_length):  # pylint: disable=redefined-outer-name
    """
    Test that the GRO reader raises an exception if the number of atoms is not
    consistent.
    """
    with pytest.raises(ValueError):
        gro.read_gro(gro_wrong_length)


def test_write_gro(gro_reference, tmpdir):
    """
    Test writing GRO file.
    """
    filename, molecule = gro_reference
    system = vermouth.System()
    system.molecules.append(molecule)
    outname = tmpdir / 'out_test.gro'
    gro.write_gro(
        system,
        outname,
        box=(10.0, 11.1, 12.2),
        title='Just a title',
    )
    DeferredFileWriter().write()
    with open(str(filename)) as ref, open(str(outname)) as out:
        assert out.read() == ref.read()



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that assigns a `mass` attribute to every node in a
molecule based on it's element.
"""


from .processor import Processor

# TODO: make the masses part of the forcefield
ATOM_MASSES = {'H': 1, 'C': 12, 'N': 14, 'O': 16, 'S': 32, 'P': 31, 'M': 0}


def attach_mass(molecule, attribute='mass'):
    """
    For every atom in `molecule` look up it's element in ``ATOM_MASSES``, and
    assign that value to `attribute`.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule to process. Is modified in-place.
    attribute: collections.abc.Hashable
        The attribute the mass is assigned to.
    """
    for node in molecule.nodes.values():
        node[attribute] = ATOM_MASSES[node['element']]


class AttachMass(Processor):
    def __init__(self, attribute='mass'):
        super().__init__()
        self.attribute = attribute

    def run_molecule(self, molecule):
        attach_mass(molecule, attribute=self.attribute)
        return molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a string formatter that can not only pad strings to a specified
length if they're too short, but also truncate them if they're too long.
"""


import string
import re
from collections import namedtuple

FormatSpec = namedtuple('FormatSpec', 'fill align sign alt zero_padding width comma decimal precision type')


class TruncFormatter(string.Formatter):
    """
    Adds the 't' option to the format specification mini-language at the end of
    the format string. If provided, the produced formatted string will be
    truncated to the specified length.
    """
    # https://stackoverflow.com/questions/44551535/access-the-cpython-string-format-specification-mini-language-parser
    format_spec_re = r'(([\s\S])?([<>=\^]))?([\+\- ])?(#)?(0)?(\d*)?(,)?((\.)(\d*))?([sbcdoxXneEfFgGn%])?'
    format_spec_re = re.compile(format_spec_re)

    def format_field(self, value, format_spec):
        """
        Implements the 't' option to truncate strings that are too long to the
        required width.

        Parameters
        ----------
        value
            The object to format.
        format_spec: str
            The format_spec describing how `value` should be formatted

        Returns
        str
            `value` formatted as per `format_spec`
        """
        if format_spec.endswith('t'):
            truncate = True
            format_spec = format_spec[:-1]
        else:
            truncate = False
        result = super().format_field(value, format_spec)
        # From here on we know the format spec is valid
        spec = FormatSpec(*self.format_spec_re.fullmatch(format_spec).group(2, 3, 4, 5, 6, 7, 8, 10, 11, 12))
        if spec.width:
            spec = spec._replace(width=int(spec.width))
        else:
            spec = spec._replace(width=0)
        if not truncate or spec.width == 0 or len(result) <= spec.width:
            return result
        # skip groups not interested in
        if not spec.type:
            if isinstance(value, str):
                spec = spec._replace(type='s')
            elif isinstance(value, int):
                spec = spec._replace(type='d')
            elif isinstance(value, float):
                spec = spec._replace(type='g')

        if not spec.align:
            if spec.type in 's':
                spec = spec._replace(align='<')
            elif spec.type in 'bcdoxXn' or spec.type in 'eEfFgGn%':
                spec = spec._replace(align='>')

        # We know len(result) > width. So there's no fill characters.
        # We also have at least width, type and align at this point.
        # We should probably do something special when it's a number with a
        # magic formatting prefix (0b, 0o, 0x) or if it has a sign. Idem for
        # exponent notation. Maybe, for numerical types we should round instead
        # of truncate the string.
        overflow = len(result) - spec.width
        if spec.align == '<':  # left chars most significant. e.g. str
            result = result[:-overflow]
        elif spec.align == '>':  # right characters most significant. e.g. int
            result = result[overflow:]
        elif spec.align == '=':  # padding between sign and digits +0000120
            # Note that this is the default for fill character 0
            raise NotImplementedError
        elif spec.align == '^':  # centered
            result = result[overflow//2:-overflow//2]

        return result


# if __name__ == '__main__':
#     formatter = TruncFormatter()
#
#     str_data = 'abcde'
#
#     print(formatter.format('{}', str_data))
#
#     print(formatter.format('"{:4.4}"', str_data))
#     print(formatter.format('"{:>4.4}"', str_data))
#     print(formatter.format('"{:4t}"', str_data))
#     print(formatter.format('"{:>4t}"', str_data))
#     print(formatter.format('"{:5t}"', str_data))
#     print(formatter.format('"{:6t}"', str_data))
#
#     int_data = 123456789
#
#     print(formatter.format('"{:4}"', int_data))
#     print(formatter.format('"{:>4}"', int_data))
#     print(formatter.format('"{:4t}"', int_data))
#     print(formatter.format('"{:>4t}"', int_data))
#     print(formatter.format('"{:<4t}"', int_data))
#     print(formatter.format('"{:^4t}"', int_data))
#     print(formatter.format('"{:5t}"', int_data))
#     print(formatter.format('"{:6t}"', int_data))
#
#     print(formatter.format('"{:11t}"', int_data))
#     print(formatter.format('"{:<11t}"', int_data))
#     print(formatter.format('"{:^11t}"', int_data))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the DoMapping processor
"""

from collections import defaultdict
import pprint

import networkx as nx
import pytest

from vermouth.processors.do_mapping import do_mapping, cover, modification_matches, apply_mod_mapping
import vermouth.forcefield
from vermouth.molecule import Molecule, Block, Link, Interaction
from vermouth.map_parser import Mapping
from vermouth.tests.helper_functions import equal_graphs

# Pylint does *not* like pytest fixtures
# pylint: disable=redefined-outer-name


FF_MARTINI = vermouth.forcefield.get_native_force_field(name='martini22')
FF_UNIVERSAL = vermouth.forcefield.get_native_force_field(name='universal')

AA_MOL = Molecule(force_field=FF_UNIVERSAL)
AA_MOL.add_nodes_from((
    (0, {'resid': 1, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
    (1, {'resid': 1, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
    (2, {'resid': 1, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
    (3, {'resid': 2, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
    (4, {'resid': 2, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
    (5, {'resid': 2, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
    (6, {'resid': 3, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
    (7, {'resid': 3, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
    (8, {'resid': 3, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
))
AA_MOL.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4), (4, 5), (4, 6), (6, 7), (7, 8)])
block_aa = Block(force_field=FF_UNIVERSAL)
block_aa.name = 'IPO'
block_aa.add_nodes_from((
    ('C1', {'resid': 1, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C2', {'resid': 1, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C3', {'resid': 1, 'resname': 'IPO', 'atomname': 'C3'}),
))
block_aa.add_edges_from([('C1', 'C2'), ('C2', 'C3')])

block_cg = Block(force_field=FF_MARTINI)
block_cg.name = 'IPO'
block_cg.add_nodes_from((('B1', {'resid': 1, 'resname': 'IPO', 'atomname': 'B1'}),))

FF_MARTINI.blocks['IPO'] = block_cg
FF_UNIVERSAL.blocks['IPO'] = block_aa


block_aa = Block(force_field=FF_UNIVERSAL)
block_aa.name = 'IPO_large'
block_aa.add_nodes_from((
    ('C1', {'resid': 1, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C2', {'resid': 1, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C3', {'resid': 1, 'resname': 'IPO', 'atomname': 'C3'}),
    ('C4', {'resid': 2, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C5', {'resid': 2, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C6', {'resid': 2, 'resname': 'IPO', 'atomname': 'C3'}),
    ('C7', {'resid': 3, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C8', {'resid': 3, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C9', {'resid': 3, 'resname': 'IPO', 'atomname': 'C3'}),
))
block_aa.add_edges_from([('C1', 'C2'), ('C2', 'C3'), ('C2', 'C4'),
                         ('C4', 'C5'), ('C5', 'C6'), ('C5', 'C7'),
                         ('C7', 'C8'), ('C8', 'C9'),
                         ])

block_cg = Block(force_field=FF_MARTINI)
block_cg.name = 'IPO_large'
block_cg.add_nodes_from((('B1', {'resid': 1, 'resname': 'IPO_large', 'atomname': 'B1'}),
                         ('B2', {'resid': 2, 'resname': 'IPO_large', 'atomname': 'B1'}),
                         ('B3', {'resid': 3, 'resname': 'IPO_large', 'atomname': 'B1'}),))
block_cg.add_edges_from([('B1', 'B2'), ('B2', 'B3')])

FF_MARTINI.blocks['IPO_large'] = block_cg
FF_UNIVERSAL.blocks['IPO_large'] = block_aa


def test_no_residue_crossing():
    """
    Make sure we don't cross residue boundaries
    """
    mapping = {'C1': {'B1': 1}, 'C2': {'B1': 1}, 'C3': {'B1': 1}}
    extra = ()
    mappings = {'universal': {'martini22': {'IPO': Mapping(FF_UNIVERSAL.blocks['IPO'],
                                                           FF_MARTINI.blocks['IPO'],
                                                           mapping=mapping,
                                                           references={},
                                                           ff_from=FF_UNIVERSAL,
                                                           ff_to=FF_MARTINI,
                                                           names=('IPO',),
                                                           extra=extra)}}}

    cg = do_mapping(AA_MOL, mappings, FF_MARTINI, attribute_keep=['chain'])

    expected = Molecule(force_field=FF_MARTINI)
    expected.add_nodes_from((
        (0, {'resid': 1, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (1, {'resid': 2, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 2}),
        (2, {'resid': 3, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 3}),
    ))
    expected.add_edges_from(([0, 1], [1, 2]))

    print(cg.nodes(data=True))
    print(cg.edges())
    print('-'*80)
    print(expected.nodes(data=True))
    print(expected.edges())
    assert equal_graphs(cg, expected)


def test_residue_crossing():
    '''
    Make sure we do cross residue boundaries and can rename residues
    '''
    mapping = {'C1': {'B1': 1}, 'C2': {'B1': 1}, 'C3': {'B1': 1},
               'C4': {'B2': 1}, 'C5': {'B2': 1}, 'C6': {'B2': 1},
               'C7': {'B3': 1}, 'C8': {'B3': 1}, 'C9': {'B3': 1},
               }
    extra = ()
    mappings = {
        'universal': {
            'martini22': {
                'IPO_large': Mapping(FF_UNIVERSAL.blocks['IPO_large'],
                                     FF_MARTINI.blocks['IPO_large'],
                                     mapping=mapping,
                                     references={},
                                     ff_from=FF_UNIVERSAL,
                                     ff_to=FF_MARTINI,
                                     names=('IPO', 'IPO', 'IPO'),
                                     extra=extra)
            }
        }
    }

    cg = do_mapping(AA_MOL, mappings, FF_MARTINI, attribute_keep=('chain',))

    expected = Molecule()
    expected.add_nodes_from((
        (0, {'resid': 1, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (1, {'resid': 2, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (2, {'resid': 3, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
    ))
    expected.add_edges_from(([0, 1], [1, 2]))

    print(cg.nodes(data=True))
    print(cg.edges())
    print('-'*80)
    print(expected.nodes(data=True))
    print(expected.edges())
    assert equal_graphs(cg, expected)


def _map_weights(mapping):
    """
    Get the weights associated with a mapping
    """
    inv_map = defaultdict(list)
    for from_, tos in mapping.items():
        for to in tos:
            inv_map[to].append(from_)
    weights = {}
    for bd, atoms in inv_map.items():
        weights[bd] = {atom: 1 for atom in atoms}
    return weights


def test_peptide():
    """
    Test multiple cg beads in residue
    """
    gly = {'C': {'BB': 1}, 'N': {'BB': 1}, 'O': {'BB': 1}, 'CA': {'BB': 1}}
    ile = {'C': {'BB': 1}, 'N': {'BB': 1}, 'O': {'BB': 1}, 'CA': {'BB': 1},
           'CB': {'SC1': 1}, 'CG1': {'SC1': 1}, 'CG2': {'SC1': 1}, 'CD': {'SC1': 1}}
    leu = {'C': {'BB': 1}, 'N': {'BB': 1}, 'O': {'BB': 1}, 'CA': {'BB': 1},
           'CB': {'SC1': 1}, 'CG': {'SC1': 1}, 'CD1': {'SC1': 1}, 'CD2': {'SC1': 1}}
    extra = ()

    mappings = {'universal': {'martini22': {}}}
    mappings['universal']['martini22']['GLY'] = Mapping(FF_UNIVERSAL.blocks['GLY'],
                                                        FF_MARTINI.blocks['GLY'],
                                                        mapping=gly,
                                                        references={},
                                                        ff_from=FF_UNIVERSAL,
                                                        ff_to=FF_MARTINI,
                                                        names=('GLY',),
                                                        extra=extra)
    mappings['universal']['martini22']['ILE'] = Mapping(FF_UNIVERSAL.blocks['ILE'],
                                                        FF_MARTINI.blocks['ILE'],
                                                        mapping=ile,
                                                        references={},
                                                        ff_from=FF_UNIVERSAL,
                                                        ff_to=FF_MARTINI,
                                                        names=('ILE',),
                                                        extra=extra)
    mappings['universal']['martini22']['LEU'] = Mapping(FF_UNIVERSAL.blocks['LEU'],
                                                        FF_MARTINI.blocks['LEU'],
                                                        mapping=leu,
                                                        references={},
                                                        ff_from=FF_UNIVERSAL,
                                                        ff_to=FF_MARTINI,
                                                        names=('LEU',),
                                                        extra=extra)

    peptide = Molecule(force_field=FF_UNIVERSAL)
    aa = FF_UNIVERSAL.blocks['GLY'].to_molecule()
    peptide.merge_molecule(aa)

    aa = FF_UNIVERSAL.blocks['ILE'].to_molecule()
    peptide.merge_molecule(aa)

    aa = FF_UNIVERSAL.blocks['LEU'].to_molecule()
    peptide.merge_molecule(aa)

    peptide.add_edge(list(peptide.find_atoms(atomname='N', resid=1))[0],
                     list(peptide.find_atoms(atomname='C', resid=2))[0])
    peptide.add_edge(list(peptide.find_atoms(atomname='N', resid=2))[0],
                     list(peptide.find_atoms(atomname='C', resid=3))[0])

    for node in peptide:
        peptide.nodes[node]['atomid'] = node + 1
        peptide.nodes[node]['chain'] = ''

    cg = do_mapping(peptide, mappings, FF_MARTINI, attribute_keep=('chain',))

    expected = Molecule(force_field=FF_MARTINI)
    expected.add_nodes_from({1: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 1,
                                 'resid': 1,
                                 'resname': 'GLY'},
                             2: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 2,
                                 'resid': 2,
                                 'resname': 'ILE'},
                             3: {'atomname': 'SC1',
                                 'atype': 'AC1',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 3,
                                 'resid': 2,
                                 'resname': 'ILE'},
                             4: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 4,
                                 'resid': 3,
                                 'resname': 'LEU'},
                             5: {'atomname': 'SC1',
                                 'atype': 'AC1',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 5,
                                 'resid': 3,
                                 'resname': 'LEU'}}.items()
                            )
    expected.add_edges_from([(1, 2), (2, 3), (2, 4), (4, 5)])

    for node in expected:
        expected.nodes[node]['atomid'] = node + 1
    print(cg.nodes(data=True))
    print(cg.edges())
    print('-'*80)
    print(expected.nodes(data=True))
    print(expected.edges())

    assert equal_graphs(cg, expected)


@pytest.mark.parametrize('to_cover, options, expected', (
    ([], [], []),
    ([], [[1, 2], [3, 4]], []),
    ([1, 2, 3], [[1, 2, 3]], [[1, 2, 3]]),
    ([1, 2, 3], [[1], [2], [3]], [[1], [2], [3]]),
    ([1, 2, 3], [[1, 2, 3], [4]], [[1, 2, 3]]),
    ([1, 2, 3], [[1], [2], [3], [4]], [[1], [2], [3]]),
    ([1, 2, 3], [[4], [1, 2, 3]], [[1, 2, 3]]),
    ([1, 2, 3], [[4], [1], [2], [3], [5]], [[1], [2], [3]]),
    ([1, 2, 3], [[1, 3], [1], [2], [3], [4]], [[1, 3], [2]]),
    ([1, 2, 3], [[1, 3, 4], [1], [2], [3], [4]], [[1], [2], [3]]),
    ([1, 2, 3], [[1, 3, 4], [1], [2], [3], [1, 2, 3]], [[1], [2], [3]]),
    ([1, 2, 3], [], None),
    ([1, 2, 3], [[1, 2]], None),
    ([1, 2, 3], [[1, 2], [2, 3]], None),
    ([1, 2, 3], [[1, 2], [3, 4]], None),
    ([1, 2, 3], [[1, 2], [4]], None),
))
def test_cover(to_cover, options, expected):
    """
    Test the cover function
    """
    output = cover(to_cover, options)
    assert output == expected


@pytest.fixture
def modifications():
    """
    Provides modifications
    """
    mods = {}
    mod_a = Link(force_field=FF_UNIVERSAL, name='mA')
    mod_a.add_node('mA', atomname='mA', PTM_atom=True, modifications=[mod_a])
    mods['mA'] = mod_a

    mod_c = Link(force_field=FF_UNIVERSAL, name='mC')
    mod_c.add_node('mC', atomname='mC', PTM_atom=True, modifications=[mod_c])
    mods['mC'] = mod_c

    mod_d = Link(force_field=FF_UNIVERSAL, name='mD')
    mod_d.add_node('mD', atomname='mD', PTM_atom=True, modifications=[mod_d])
    mods['mD'] = mod_d

    mod_fg = Link(force_field=FF_UNIVERSAL, name='mFG')
    mod_fg.add_node('mF', atomname='mF', PTM_atom=True, modifications=[mod_fg])
    mod_fg.add_node('mG', atomname='mG', PTM_atom=True, modifications=[mod_fg])
    mod_fg.add_edge('mF', 'mG')
    mod_fg.add_interaction('bond', ['mF', 'mG'], (3, 4))
    mods['mFG'] = mod_fg

    mod_i = Link(force_field=FF_UNIVERSAL, name='mI')
    mod_i.add_node('mI', atomname='mI', PTM_atom=True, modifications=[mod_i])
    mods['mI'] = mod_i
    mod_i2 = Link(force_field=FF_UNIVERSAL, name='mI2')
    mod_i2.add_node('mI2', atomname='mI2', PTM_atom=True, modifications=[mod_i2])
    mods['mI2'] = mod_i2

    mod_j = Link(force_field=FF_UNIVERSAL, name='mJ')
    mod_j.add_node('mJ', atomname='mJ', PTM_atom=True, modifications=[mod_j])
    mods['mJ'] = mod_j
    mod_j2 = Link(force_field=FF_UNIVERSAL, name='mJ2')
    mod_j2.add_node('mJ2', atomname='mJ2', PTM_atom=True, modifications=[mod_j2])
    mods['mJ2'] = mod_j2
    mod = Link(name=('mJ', 'mJ2'), force_field=FF_UNIVERSAL)
    mod.add_nodes_from((['mJ', {'atomname': 'mJ', 'PTM_atom': True, 'modifications': [mods['mJ']]}],
                        ['mJ2', {'atomname': 'mJ2', 'PTM_atom': True, 'modifications': [mods['mJ2']]}],
                        ['J', {'atomname': 'J', 'PTM_atom': False}]))
    mod.add_edges_from([('J', 'mJ'), ('J', 'mJ2')])
    mods[('mJ', 'mJ2')] = mod
    return mods


@pytest.fixture
def modified_molecule(modifications):
    """
    Provides a molecule with modifications
    """
    mol = Molecule(force_field=FF_UNIVERSAL)
    mol.add_nodes_from(enumerate((
        # Lone PTM
        {'atomname': 'A', 'resid': 1, 'modifications': [modifications['mA']]},
        {'atomname': 'mA', 'resid': 1, 'PTM_atom': True, 'modifications': [modifications['mA']]},
        {'atomname': 'B', 'resid': 2},  # Spacer
        # Two PTMs on neighbouring residues
        {'atomname': 'C', 'resid': 3, 'modifications': [modifications['mC']]},
        {'atomname': 'mC', 'resid': 3, 'PTM_atom': True, 'modifications': [modifications['mC']]},
        {'atomname': 'D', 'resid': 4, 'modifications': [modifications['mD']]},
        {'atomname': 'mD', 'resid': 4, 'PTM_atom': True, 'modifications': [modifications['mD']]},
        {'atomname': 'E', 'resid': 5},  # Spacer
        # Bridging PTMs
        {'atomname': 'F', 'resid': 6, 'modifications': [modifications['mFG']]},
        {'atomname': 'mF', 'resid': 6, 'PTM_atom': True, 'modifications': [modifications['mFG']]},
        {'atomname': 'G', 'resid': 7, 'modifications': [modifications['mFG']]},
        {'atomname': 'mG', 'resid': 7, 'PTM_atom': True, 'modifications': [modifications['mFG']]},
        {'atomname': 'H', 'resid': 8},  # Spacer
        # Two PTMs for one residue
        {'atomname': 'I', 'resid': 9, 'modifications': [modifications['mI'], modifications['mI2']]},
        {'atomname': 'mI', 'resid': 9, 'PTM_atom': True, 'modifications': [modifications['mI']]},
        {'atomname': 'mI2', 'resid': 9, 'PTM_atom': True, 'modifications': [modifications['mI2']]},
        # Two PTMs for one residue, but a single mod mapping
        {'atomname': 'J', 'resid': 10, 'modifications': [modifications['mJ'], modifications['mJ2']]},
        {'atomname': 'mJ', 'resid': 10, 'PTM_atom': True, 'modifications': [modifications['mJ']]},
        {'atomname': 'mJ2', 'resid': 10, 'PTM_atom': True, 'modifications': [modifications['mJ2']]}
    )))
    mol.add_edges_from((
        (0, 1), (0, 2),  # A
        (2, 3),  # B
        (3, 4), (3, 5),  # C
        (5, 6), (5, 7),  # D
        (7, 8),  # E
        (8, 9), (8, 10),  # F
        (9, 11),  # Bridge between mF and mG
        (10, 11), (10, 12),  # G
        (12, 13),  # H
        (13, 14), (13, 15), (13, 16),  # I
        (16, 17), (16, 18)  # J
    ))
    return mol


def test_mod_matches(modified_molecule, modifications):
    """
    Test modification matches
    """
    mappings = []
    for name in 'ABCDEFGHIJ':
        block = Block(force_field=FF_UNIVERSAL, name=name)
        block.add_node(name, atomname=name)
        mappings.append(Mapping(block, block, mapping={name: {name: 1}},
                                references={}, ff_from=FF_UNIVERSAL,
                                ff_to=FF_UNIVERSAL, names=(name,)))
    for mod in modifications.values():
        name = (mod.name,) if isinstance(mod.name, str) else mod.name
        mappings.append(Mapping(mod, mod, mapping={idx: {idx: 1} for idx in mod},
                                references={}, ff_from=FF_UNIVERSAL,
                                ff_to=FF_UNIVERSAL, names=name, type='modification'))

    print([m.names for m in mappings])
    found = list(modification_matches(modified_molecule, mappings))
    expected = [
        ({1: {'mA': 1}}, modifications['mA'], {}),
        ({4: {'mC': 1}}, modifications['mC'], {}),
        ({6: {'mD': 1}}, modifications['mD'], {}),
        ({9: {'mF': 1}, 11: {'mG': 1}}, modifications['mFG'], {}),
        ({14: {'mI': 1}}, modifications['mI'], {}),
        ({15: {'mI2': 1}}, modifications['mI2'], {}),
        ({16: {'J': 1}, 17: {'mJ': 1}, 18: {'mJ2': 1}}, modifications['mJ', 'mJ2'], {}),
    ]
    pprint.pprint(found)
    pprint.pprint(expected)
    print([e[1].name for e in expected])
    print([e[1].name for e in found])
    assert len(expected) == len(found)
    e_matches, e_mods, e_refs = zip(*expected)
    f_matches, f_mods, f_refs = zip(*found)

    for f_match in f_matches:
        assert f_match in e_matches
    for f_ref in f_refs:
        assert f_ref in e_refs
    e_mod_names = [(mod.name,) if isinstance(mod.name, str) else mod.name for mod in e_mods]
    for f_mod in f_mods:
        assert f_mod.name in e_mod_names


def test_apply_mod_mapping(modified_molecule, modifications):
    """
    Test apply_mod_mapping
    """
    graph_out = Molecule(force_field=FF_UNIVERSAL)
    graph_out.add_nodes_from([
        (0, {'atomname': 'A', 'resid': 1})
    ])
    mol_to_out = {0: {0: 1}}
    out_to_mol = {0: {0: 1}}
    match = ({1: {'mA': 1}}, modifications['mA'], {})

    out = apply_mod_mapping(match, modified_molecule, graph_out, mol_to_out, out_to_mol)
    print(mol_to_out)
    print(out_to_mol)
    print(graph_out.nodes[1])
    print(modifications['mA'].nodes['mA'])
    for key in modifications['mA'].nodes['mA']:
        assert graph_out.nodes[1][key] == modifications['mA'].nodes['mA'][key]
    assert out == ({}, {})
    assert mol_to_out == {0: {0: 1}, 1: {1: 1}}
    assert out_to_mol == {0: {0: 1}, 1: {1: 1}}

    graph_out.add_node(2, atomname='J', resid=2)
    mol_to_out[16] = {2: 1}
    out_to_mol[2] = {16: 1}

    out = apply_mod_mapping((
        {16: {'J': 1}, 17: {'mJ': 1}, 18: {'mJ2': 1}},
        modifications['mJ', 'mJ2'], {}
    ), modified_molecule, graph_out, mol_to_out, out_to_mol)
    print(mol_to_out)
    print(out_to_mol)
    assert out == ({}, {})
    assert mol_to_out == {0: {0: 1}, 1: {1: 1}, 16: {2: 1}, 17: {3: 1}, 18: {4: 1}}
    assert out_to_mol == {0: {0: 1}, 1: {1: 1}, 2: {16: 1}, 3: {17: 1}, 4: {18: 1}}


def test_do_mapping_mods(modified_molecule, modifications):
    """
    Test do_mapping on a molecule with modifications
    """
    mappings = {}
    for name in 'ABCDEFGHIJ':
        block = Block(force_field=FF_UNIVERSAL, name=name)
        block.add_node(name, atomname=name, resid=1)
        mappings[name] = Mapping(block, block, mapping={name: {name: 1}},
                                 references={}, ff_from=FF_UNIVERSAL,
                                 ff_to=FF_UNIVERSAL, names=(name,))
    for mod in modifications.values():
        name = (mod.name,) if isinstance(mod.name, str) else mod.name
        mappings[name] = Mapping(mod, mod, mapping={idx: {idx: 1} for idx in mod},
                                 references={}, ff_from=FF_UNIVERSAL,
                                 ff_to=FF_UNIVERSAL, names=name, type='modification')
    mappings = {'universal': {'universal': mappings}}

    out = do_mapping(modified_molecule, mappings, FF_UNIVERSAL,
                     attribute_keep=('chain', 'resid'))
    pprint.pprint(list(out.nodes(data=True)))
    pprint.pprint(list(out.edges))

    expected = modified_molecule.copy()
    for node in expected.nodes:
        expected.nodes[node]['mapping_weights'] = {node: 1}
        expected.nodes[node]['graph'] = expected.subgraph([node])

    expected = nx.relabel_nodes(expected, {idx: idx+1 for idx in expected})
    pprint.pprint(list(expected.nodes(data='atomname')))
    pprint.pprint(list(expected.edges))

    assert equal_graphs(expected, out, node_attrs=['atomname', 'resid', 'mapping_weights'])
    assert out.interactions['bond'] == [Interaction(atoms=(10, 11), parameters=(3, 4), meta={})]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that annotates a molecule with desired mutations and
modifications.
"""

from .processor import Processor
from ..log_helpers import StyleAdapter, get_logger
from ..graph_utils import make_residue_graph
LOGGER = StyleAdapter(get_logger(__name__))


def parse_residue_spec(resspec):
    """
    Parse a residue specification: [<chain>-][<resname>][[#]<resid>] where
    resid is /[0-9]+/.
    If resname ends in a number and a resid is also specified, the # separator
    is required.
    Returns a dictionary with keys 'chain', 'resname', and 'resid' for the
    fields that are specified. Resid will be an int.

    Parameters
    ----------
    resspec: str

    Returns
    -------
    dict
    """
    # A-LYS2 or PO4#2
    # <chain>-<resname><resid>
    *chain, res = resspec.split('-', 1)
    res, *resid = res.rsplit('#', 1)
    if resid:  # [] if False
        resname = res
        resid = resid[0]
    else:
        idx = 0
        for idx, char in reversed(list(enumerate(res))):
            if not char.isdigit():
                idx += 1
                break
        resname = res[:idx]
        resid = res[idx:]

    out = {}
    if resid:
        resid = int(resid)
        out['resid'] = resid
    if resname:
        out['resname'] = resname
    if chain:
        out['chain'] = chain[0]
    return out


def _subdict(dict1, dict2):
    """True if dict1 <= dict2
    All items in dict1 must be in dict2.
    """
    for key, val in dict1.items():
        if key not in dict2 or dict2[key] != val:
            return False
    return True


def residue_matches(resspec, residue_graph, res_idx):
    """
    Returns True iff resspec describes residue_graph.nodes[res_idx]. The
    'resname's nter and cter match the residues with a degree of 1 and
    with the lowest and highest residue numbers respectively.

    Parameters
    ----------
    resspec: dict
        Attributes that must be present in the residue node. 'resname' is
        treated specially as described above.
    residue_graph: networkx.Graph
        A graph with one node per residue.
    res_idx: collections.abc.Hashable
        A node index in residue_graph.

    Returns
    -------
    bool
        Whether resspec describes the node res_idx in residue_graph.
    """
    res_node = residue_graph.nodes[res_idx]
    residue = {key: res_node.get(key)
               for key in 'chain resid resname insertion_code'.split()}
    # Find all residues with degree 1: the ones with a resid lower than
    # their neighbour will be Nter, those with a resid higher than their
    # neighbour Cter
    # FIXME: Once residue_graph is a digraph we can do something much much
    #        more clever, addressing arbitrarily branched polymers and
    #        termini
    if residue_graph.degree[res_idx] == 1 and resspec.get('resname') in ['nter', 'cter']:
        if not _terminal_matches(resspec.get('resname'), residue_graph, res_idx):
            return False
        # Remove resname and resid from the resspec, since they have special meaning
        # in this case.
        resspec = resspec.copy()
        resspec.pop('resname', '')
        resspec.pop('resid', 0)
    return _subdict(resspec, residue)


def _terminal_matches(resname, residue_graph, res_idx):
    """
    Tests whether the node res_idx in graph residue_graph is a cter or nter, as
    determined by resname. Resname must be one of nter or cter.
    It is assumed that the degree of the specified node is 1.

    Parameters
    ----------
    resname: str
        Must be either nter or cter
    residue_graph: networkx.Graph
    res_idx: collections.abc.Hashable
        Key in residue_graph

    Returns
    -------
    bool
        If resname is nter, returns True iff the resid of the specified node is
        lower than that of its neighbour. If resname is cter, returns True iff
        the resid of the specified node is larger than that of its neighbour.

    Raises
    ------
    KeyError
        If resname is neither cter nor nter
    """
    neighbour = list(residue_graph[res_idx])[0]  # Only one neighbour by definition.
    resid = residue_graph.nodes[res_idx].get('resid', 0)
    neighbour_resid = residue_graph.nodes[neighbour].get('resid', 0)
    if resname == 'nter':
        return resid < neighbour_resid
    elif resname == 'cter':
        return resid > neighbour_resid
    raise KeyError('Unknown residue name in provided resspec. Found {}, only '
                   'cter and nter are known'.format(resname))


def _format_resname(res):
    """
    Provisional function that performs the opposite of parse_residue_spec.
    Poorly tested, use at own risk.
    """
    chain = res.get('chain', '')
    out = ''
    if chain:
        out += chain + '-'
    resname = res.get('resname')
    out += resname
    if resname and resname[-1].isdigit():
        out += '#'
    out += str(res.get('resid', ''))
    out += res.get('insertion_code', '')
    return out


def annotate_modifications(molecule, modifications, mutations):
    """
    Annotate nodes in molecule with the desired modifications and mutations

    Parameters
    ----------
    molecule: networkx.Graph
    modifications: list[tuple[dict, str]]
        The modifications to apply. The first element is a dictionary contain
        the attributes a residue has to fulfill. It can contain the elements
        'chain', 'resname' and 'resid'. The second element is the modification
        that should be applied.
    mutations: list[tuple[dict, str]]
        The mutations to apply. The first element is a dictionary contain
        the attributes a residue has to fulfill. It can contain the elements
        'chain', 'resname' and 'resid'. The second element is the mutation that
        should be applied.

    Raises
    ------
    NameError
        When a modification is not recognized.
    """
    if not modifications and not mutations:
        return

    # We need to do very similar but not quite identical things for
    # modifications and mutations. Associate them with the correct name and  FF
    # elements.
    associations = [(modifications, 'modification', molecule.force_field.modifications),
                    (mutations, 'mutation', molecule.force_field.blocks)]

    residue_graph = make_residue_graph(molecule)
    for res_idx in residue_graph:
        for mutmod, key, library in associations:
            for resspec, mod in mutmod:
                if residue_matches(resspec, residue_graph, res_idx):
                    if mod != 'none' and mod not in library:
                        raise NameError('{} is not known as a {} for '
                                        'force field {}'
                                        ''.format(mod, key, molecule.force_field.name))
                    res = residue_graph.nodes[res_idx]
                    LOGGER.debug('Annotating {} with {} {}',
                                 _format_resname(res), key, mod)
                    for node_idx in res['graph']:
                        molecule.nodes[node_idx][key] = molecule.nodes[node_idx].get(key, []) + [mod]


class AnnotateMutMod(Processor):
    def __init__(self, modifications=None, mutations=None):
        if not modifications:
            modifications = []
        if not mutations:
            mutations = []
        self.modifications = []
        for resspec, val in modifications:
            self.modifications.append((parse_residue_spec(resspec), val))
        self.mutations = []
        for resspec, val in mutations:
            self.mutations.append((parse_residue_spec(resspec), val))

    def run_molecule(self, molecule):
        annotate_modifications(molecule, self.modifications, self.mutations)
        return molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Set of tools to add and remove edges.
"""

import functools
import numpy as np
import networkx as nx

from . import KDTree
from . import selectors
from . import geometry


def _edge_is_between_selections(edge, selection_a, selection_b):
    """
    Returns ``True`` is the edge has one end in each selection.

    Parameters
    ----------
    edge: tuple[int, int]
    selection_a: collections.abc.Container[collections.abc.Hashable]
    selection_b: collections.abc.Container[collections.abc.Hashable]

    Returns
    -------
    bool
    """
    return (
        (edge[0] in selection_a and edge[1] in selection_b)
        or (edge[1] in selection_a and edge[0] in selection_b)
    )


def prune_edges_between_selections(molecule, selection_a, selection_b):
    """
    Remove edges which have their ends part of given selections.

    An edge is removed if has one end that is part of 'selection_a', and the
    other end part of 'selection_b'.

    Parameters
    ----------
    molecule: networkx.Graph
        Molecule to prune in-place.
    selection_a: collections.abc.Iterable[collections.abc.Hashable]
        List of node keys from the molecule.
    selection_b: collections.abc.Iterable[collections.abc.Hashable]
        List of node keys from the molecule.

    See Also
    --------
    prune_edges_with_selectors
    """
    selection_a = set(selection_a)
    selection_b = set(selection_b)
    # to_prune is the list of edges to remove. It cannot be a generator or
    # removing the edges will be modifying the datastructure on which we
    # iterate.
    to_prune = [
        edge for edge in molecule.edges
        if _edge_is_between_selections(edge, selection_a, selection_b)
    ]
    molecule.remove_edges_from(to_prune)


def prune_edges_with_selectors(molecule, selector_a, selector_b=None):
    """
    Remove edges with the ends between selections defined by selectors.

    An edge is removed if one of its end is part of the selection defined by
    'selector_a', and its other end is part of the selection defined by
    'selector_b'. A selector is a function that accept a node dictionary as
    argument and returns ``True`` if the node is part of the selection.

    The 'selection_b' argment is optional. If it is ``None``, then 'selector_a'
    is used for the selection at both ends.

    Parameters
    ----------
    molecule: networkx.Graph
        Molecule to prune in-place.
    selector_a: collections.abc.Callable
        A selector for one end of the edges.
    selector_b: collections.abc.Callable
        A selector for the second end of the edges. If set to ``None``, then
        'selector_a' is used for both ends.

    See Also
    --------
    prune_edges_between_selections
    """
    selection_a = selectors.filter_minimal(molecule, selector_a)
    if selector_b is None:
        selector_b = selector_a
    selection_b = selectors.filter_minimal(molecule, selector_b)
    prune_edges_between_selections(molecule, selection_a, selection_b)


def add_edges_at_distance(molecule, threshold,
                          selection_a, selection_b, attribute='position'):
    """
    Add edges within a molecule when the distance is below a threshold.

    Create edges within a molecule between nodes that have an end part of
    'selection_a', the other end part of 'selection_b', and a distance between
    the ends that is lesser than the given threshold.

    All nodes that are part of 'selection_a' or 'selection_b' must have a
    position stored under the attribute which key is given with the 'attribute'
    argument. That key is 'position' by default. If at least one node has the
    position missing, then a :exc:`KeyError` is raised.

    Parameters
    ----------
    molecule: networkx.Graph
        Molecule to modify in-place.
    threshold: float
        The distance threshold under which edges will be created. The distance
        is expressed in nm.
    selection_a: collections.abc.Iterable[collections.abc.Hashable]
        List of node keys from the molecule.
    selection_b: collections.abc.Iterable[collections.abc.Hashable]
        List of node keys from the molecule.
    attribute: collections.abc.Hashable
        Name of the key in the node dictionaries under which the coordinates
        are stored.

    Raises
    ------
    KeyError
        At least one node from the selections does not have a position.
    """
    selection_a = set(selection_a)
    selection_b = set(selection_b)
    keys_a = np.array([key for key in molecule.nodes.keys() if key in selection_a])
    keys_b = np.array([key for key in molecule.nodes.keys() if key in selection_b])
    coordinates_a = np.stack([
        molecule.nodes[key][attribute] for key in keys_a
    ])
    coordinates_b = np.stack([
        molecule.nodes[key][attribute] for key in keys_b
    ])

    distance_matrix = geometry.distance_matrix(coordinates_a, coordinates_b)
    index_a, index_b = np.where(distance_matrix < threshold)
    edges = (
        (node1, node2, {'distance': distance})
        for node1, node2, distance
        in zip(keys_a[index_a], keys_b[index_b], distance_matrix[index_a, index_b])
    )

    molecule.add_edges_from(edges)


def add_inter_molecule_edges(molecules, edges):
    """
    Create edges between molecules.

    The function is given a list of molecules and a list of edges. Each edge is
    provided as a tuple of two nodes, each node being a tuple of the molecule
    index in the list of molecule, and the node key in that molecule. An edge
    therefore looks like ``((0, 10), (2, 20))`` where ``1`` and ``2`` are
    indices of molecules in `molecules`, ``10`` is the key of a node from
    ``molecules[0]``, and ``20`` is the key of a node from ``molecules[2]``.

    The function **can** create edges within a molecule if the same molecule
    index is given for both ends of edges.

    Molecules that get linked are merged. In a merged molecule, the order of
    the input molecules is kept. In a list of molecules numbered from 0 to 4,
    if molecules 1, 2, and 4 are merged, then the result molecules are, in
    order, 0, 1-2-4, 3.

    Parameters
    ----------
    molecules: collections.abc.Sequence[vermouth.molecule.Molecule]
        List of molecules to link.
    edges: collections.abc.Iterable[tuple[int, collections.abc.Hashable]]
        List of edges in a ``(molecule_index, node_key)`` format as described
        above. Edges can have a third element, it is then a dictionary of
        attributes to be attached to the edge.

    Returns
    -------
    list
        New list of molecules.
    """
    edges = tuple(edges)  # We iterate multiple time trhough it.
    molecule_graph = nx.Graph()
    molecule_graph.add_nodes_from(range(len(molecules)))
    molecule_edges = [(edge[0][0], edge[1][0]) for edge in edges]
    molecule_graph.add_edges_from(molecule_edges)
    components = nx.connected_components(molecule_graph)

    # Do the merging
    new_molecules = []
    correspondance = {}
    for new_index, component in enumerate(components):
        # Connected components are yielded as sets. We need to be able to
        # iterate over them in numerical order.
        component = sorted(component)

        # The first molecule in the component is the base for the merge.
        base_index = component[0]
        base_molecule = molecules[base_index]
        new_molecules.append(base_molecule)
        for key in base_molecule:
            correspondance[(base_index, key)] = (new_index, key)

        for other_index in component[1:]:
            other_molecule = molecules[other_index]
            mol_correspondance = base_molecule.merge_molecule(other_molecule)
            for before, after in mol_correspondance.items():
                correspondance[(other_index, before)] = (new_index, after)

    # Create the edges
    for edge in edges:
        new_edge = (correspondance[edge[0]], correspondance[edge[1]])
        assert new_edge[0][0] == new_edge[1][0], \
            'The two ends of the edge are not part of the same molecule.'
        molecule_index = new_edge[0][0]
        new_molecules[molecule_index].add_edge(new_edge[0][1], new_edge[1][1])
        if len(edge) > 2:
            in_mol_edge = (
                new_molecules[molecule_index]
                .edges[new_edge[0][1], new_edge[1][1]]
            )
            in_mol_edge.update(edge[2])

    return new_molecules


def pairs_under_threshold(molecules, threshold,
                          selection_a, selection_b,
                          attribute='position', min_edges=0):
    """
    List pairs of nodes from a selection that are closer than a threshold.

    Get the distance between nodes from multiple molecules and list the pairs
    that are closer than the given threshold. The molecules are given as a list
    of molecules, the selection is a list of nodes each of them a tuple
    ``(index of the molecule in the list, key of the node in the molecule)``.
    The result of the function is a generator of node pairs followed by the
    distance between the nodes, each node formated as in the selection.

    All nodes from the selection must have a position accessible under the key
    given as the 'attribute' argument. That key is 'position' by default.

    With the `min_edges` argument, one can prevent pairs to be selected if
    there is a path between two nodes that is shorter than a given number of
    edges.

    Parameters
    ----------
    molecules: collections.abc.Collection[vermouth.molecule.Molecule]
        A list of :class:`vermouth.molecule.Molecule`.
    threshold: float
        A distance threshold in nm. Pairs are return if the nodes are closer
        than this threshold.
    selection_a: collections.abc.Iterable[collections.abc.Hashable]
        List of nodes to consider at one end of the pairs. The format is
        described above.
    selection_b: collections.abc.Iterable[collections.abc.Hashable]
        List of nodes to consider at the other end of the pairs. The format is
        described above.
    attribute: collections.abc.Hashable
        The dictionary key under which the node positions are stored in the
        nodes.
    min_edges: int
        Do not select pairs that are connected by less than that number of
        edges.

    Yields
    ------
    tuple[collections.abc.Hashable, collections.abc.Hashable, float]
        Pairs of node closer than the threshold in the format described above
        and the distance between the nodes.

    Raises
    ------
    KeyError
        Raised if a node from the selection does not have a position.

    Notes
    -----

    Symetric node pairs are not deduplicated.
    """
    coordinates_a = []
    for key in selection_a:
        coordinates_a.append(molecules[key[0]].nodes[key[1]][attribute])
    coordinates_b = []
    for key in selection_b:
        coordinates_b.append(molecules[key[0]].nodes[key[1]][attribute])
    if not coordinates_a or not coordinates_b:
        return
    kdtree_a = KDTree(coordinates_a)
    kdtree_b = KDTree(coordinates_b)
    sparse_distance_matrix = kdtree_a.sparse_distance_matrix(kdtree_b, threshold)
    for (idx, jdx), distance_between in sparse_distance_matrix.items():
        key_a = selection_a[idx]
        key_b = selection_b[jdx]
        if key_a != key_b and distance_between < threshold:
            if not _are_close(min_edges, molecules, key_a, key_b):
                yield (key_a, key_b, distance_between)


def _are_close(threshold, molecules, key_a, key_b):
    if not threshold:
        return False
    if key_a[0] != key_b[0]:
        return False
    try:
        length = nx.shortest_path_length(
            molecules[key_a[0]], key_a[1], key_b[1]
        )
    except nx.exception.NetworkXNoPath:
        return False
    else:
        return length < threshold


def select_nodes_multi(molecules, selector):
    """
    Find the nodes that correspond to a selector among multiple molecules.

    Runs a selector over multiple molecules. The selector must be a function
    that takes a node dictionary as argument and returns ``True`` if the node
    should be selected. The selection is yielded as tuples of a molecule indice
    from the molecule list input, and a key from the molecule.

    Parameters
    ----------
    molecule: collections.abc.Iterable[Molecule]
        A list of molecules.
    selector: collections.abc.Callable
        A selector function.

    Yields
    ------
    tuple[int, collections.abc.Hashable]
        Molecule/key identifier for the selected nodes.
    """
    for molecule_idx, molecule in enumerate(molecules):
        for key, node in molecule.nodes.items():
            if selector(node):
                yield (molecule_idx, key)


def add_edges_threshold(molecules, threshold,
                        templates_a, templates_b,
                        attribute='position', min_edges=0):
    """
    Add edges between two selections when under a given threshold.

    Edges are added within and between the molecules and connect nodes that
    match the given template. Molecules that get connected by an edge are
    merged and the new list of molecules is returned.

    Parameters
    ----------
    molecules: collections.abc.Sequence[Molecule]
        A list of molecules.
    threshold: float
        The distance threshold in nanometers under which an edge is created.
    templates_a: dict
        A list of templates; a node need to match at least one of them to be
        selected at one end.
    templates_b: dict
        A list of templates; a node need to match at least one of them to be
        selected at the other end.
    attribute: str
        Name of the key in the node dictionaries under which the coordinates
        are stored.
    min_edges: int
        Minimum number of edges between to nodes for an edge to be added.

    Returns
    -------
    list[vermouth.molecule.Molecule]
        A new list of molecules.
    """
    selector_a = functools.partial(
        selectors.proto_multi_templates, templates=templates_a
    )
    selection_a = list(select_nodes_multi(molecules, selector_a))
    selector_b = functools.partial(
        selectors.proto_multi_templates, templates=templates_b
    )
    selection_b = list(select_nodes_multi(molecules, selector_b))
    edges = pairs_under_threshold(molecules, threshold,
                                  selection_a, selection_b,
                                  attribute, min_edges=min_edges)
    edges = (
        (node1, node2, {'distance': distance})
        for node1, node2, distance in edges
    )
    new_molecules = add_inter_molecule_edges(molecules, edges)
    return new_molecules



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Read .ff files.

The FF file format describes molecule components for a given force field. It is
a test format devised for quick prototyping.

The format is built on top of a subset of the ITP format. Describing a block
is done in the same way an ITP file describes a molecule.
"""

import collections
import copy
import numbers
import json
from .molecule import (
    Block, Link,
    Interaction, DeleteInteraction,
    Choice, NotDefinedOrNot,
    ParamDistance, ParamAngle, ParamDihedral, ParamDihedralPhase,
)
from .parser_utils import (
    SectionLineParser, _tokenize, _substitute_macros, _parse_macro
)

# Python 3.4 does not raise JSONDecodeError but ValueError.

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

VALUE_PREDICATES = {
    'not': NotDefinedOrNot,
}

PARAMETER_EFFECTORS = {
    'dist': ParamDistance,
    'angle': ParamAngle,
    'dihedral': ParamDihedral,
    'dihphase': ParamDihedralPhase,
}


class FFDirector(SectionLineParser):
    COMMENT_CHAR = ';'
    interactions_natoms = {
        'bonds': 2,
        'angles': 3,
        'dihedrals': 4,
        'impropers': 4,
        'constraints': 2,
        'pairs': 2,
        'pairs_nb': 2,
        'SETTLE': 1,
        'virtual_sites2': 3,
        'virtual_sites3': 4,
        'virtual_sites4': 5,
        'position_restraints': 1,
        'distance_restraints': 2,
        'dihedral_restraints': 4,
        'orientation_restraints': 2,
        'angle_restraints': 4,
        'angle_restraints_z': 2
    }

    def __init__(self, force_field):
        super().__init__()
        self.force_field = force_field
        self.current_block = None
        self.current_link = None
        self.current_modification = None
        self.blocks = collections.OrderedDict()
        self.links = []
        self.modifications = []

        self.header_actions = {
            ('moleculetype', ): self._new_block,
            ('link', ): self._new_link,
            ('modification', ): self._new_modification,
        }

    def parse_header(self, line, lineno=0):
        """
        Parses a section header with line number `lineno`. Sets
        :attr:`vermouth.parser_utils.SectionLineParser.section` when applicable.
        Does not check whether `line` is a valid section header.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
            The result of calling :meth:`finalize_section`, which is called
            if a section ends.

        Raises
        ------
        KeyError
            If the section header is unknown.
        """
        
        prev_section = None

        ended = []
        section = self.section + [line.strip('[ ]').casefold()]

        if tuple(section[-1:]) in self.METH_DICT:
            prev_section = self.section
            self.section = section[-1:]

        else:
            while (tuple(section) not in self.METH_DICT
                   and len(section) > 1):
                ended.append(section.pop(-2))  # [a, b, c, d] -> [a, b, d]
            self.section = section

        result = None

        if prev_section:
            result = self.finalize_section(prev_section, ended)

        action = self.header_actions.get(tuple(self.section))
        if action:
            action()

        return result


    def finalize_section(self, previous_section, ended_section):
        """
        Called once a section is finished. It appends the current_links list
        to the links and update the block dictionary with current_block. Thereby it
        finishes the reading a given section.

        Parameters
        ---------
        previous_section: list[str]
            The last parsed section.
        ended_section: list[str]
            The sections that have been ended.
        """

        if self.current_block is not None:
            self.current_block.make_edges_from_interactions()
            self.force_field.blocks[self.current_block.name] = self.current_block

        if self.current_link is not None:
            self.current_link.make_edges_from_interactions()
            self.force_field.links.append(self.current_link)

        if self.current_modification is not None:
            self.force_field.modifications[self.current_modification.name] = self.current_modification

    def get_context(self, context_type):
        possible_contexts = {
            'block': self.current_block,
            'link': self.current_link,
            'molmeta': self.current_link,
            'modification': self.current_modification,
        }
        return possible_contexts[context_type]

    def has_context(self):
        open_contexts = [
            self.current_block, self.current_link, self.current_modification]
        return open_contexts != ([None] * len(open_contexts))

    def _new_block(self):
        self.current_block = Block(force_field=self.force_field)

    def _new_link(self):
        self.current_link = Link(force_field=self.force_field)

    def _new_modification(self):
        self.current_modification = Link(force_field=self.force_field)

    @SectionLineParser.section_parser('variables')
    def _variables(self, line, lineno=0):
        if self.has_context():
            raise IOError('The [variables] section must be defined '
                          'before the blocks, links, and modifications.')
        tokens = _tokenize(line)
        _parse_variables(tokens, self.force_field, 'variables')

    @SectionLineParser.section_parser('moleculetype')
    def _block(self, line, lineno=0):
        name, nrexcl = line.split()
        self.current_block.name = name
        self.current_block.nrexcl = int(nrexcl)

    @SectionLineParser.section_parser('moleculetype', 'atoms')
    def _block_atoms(self, line, lineno=0):
        tokens = collections.deque(_tokenize(line))
        _parse_block_atom(tokens, self.current_block)

    @SectionLineParser.section_parser('moleculetype', 'edges',
                                      negate=False, context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'non-edges',
                                      negate=True, context_type='block')
    @SectionLineParser.section_parser('link', 'edges',
                                      negate=False, context_type='link')
    @SectionLineParser.section_parser('link', 'non-edges',
                                      negate=True, context_type='link')
    @SectionLineParser.section_parser('modification', 'edges',
                                      negate=False, context_type='modification')
    def _edges(self, line, lineno=0, negate=False, context_type=''):
        context = self.get_context(context_type)
        tokens = collections.deque(_tokenize(line))
        _parse_edges(tokens, context, context_type, negate=negate)

    @SectionLineParser.section_parser('moleculetype', 'bonds', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'angles', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'impropers', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'constraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'pairs', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'exclusions', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'pairs_nb', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'SETTLE', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites2', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites3', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites4', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sitesn', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'position_restraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'distance_restraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'dihedral_restraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'orientation_restraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints', context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints_z', context_type='block')
    @SectionLineParser.section_parser('link', 'bonds', context_type='link')
    @SectionLineParser.section_parser('link', 'angles', context_type='link')
    @SectionLineParser.section_parser('link', 'impropers', context_type='link')

    @SectionLineParser.section_parser('link', 'constraints', context_type='link')
    @SectionLineParser.section_parser('link', 'pairs', context_type='link')
    @SectionLineParser.section_parser('link', 'exclusions', context_type='link')
    @SectionLineParser.section_parser('link', 'pairs_nb', context_type='block')
    @SectionLineParser.section_parser('link', 'SETTLE', context_type='link')
    @SectionLineParser.section_parser('link', 'virtual_sites2', context_type='link')
    @SectionLineParser.section_parser('link', 'virtual_sites3', context_type='link')
    @SectionLineParser.section_parser('link', 'virtual_sites4', context_type='link')
    @SectionLineParser.section_parser('link', 'virtual_sitesn', context_type='link')
    @SectionLineParser.section_parser('link', 'position_restraints', context_type='link')
    @SectionLineParser.section_parser('link', 'distance_restraints', context_type='link')
    @SectionLineParser.section_parser('link', 'dihedral_restraints', context_type='link')
    @SectionLineParser.section_parser('link', 'orientation_restraints', context_type='link')
    @SectionLineParser.section_parser('link', 'angle_restraints', context_type='link')
    @SectionLineParser.section_parser('link', 'angle_restraints_z', context_type='link')
    @SectionLineParser.section_parser('link', '!bonds', context_type='link')
    @SectionLineParser.section_parser('link', '!angles', context_type='link')
    @SectionLineParser.section_parser('link', '!impropers', context_type='link')
    @SectionLineParser.section_parser('link', '!constraints', context_type='link')
    @SectionLineParser.section_parser('link', '!pairs', context_type='link')
    @SectionLineParser.section_parser('link', '!exclusions', context_type='link')
    @SectionLineParser.section_parser('link', '!pairs_nb', context_type='link')
    @SectionLineParser.section_parser('link', '!SETTLE', context_type='link')
    @SectionLineParser.section_parser('link', '!virtual_sites2', context_type='link')
    @SectionLineParser.section_parser('link', '!virtual_sites3', context_type='link')
    @SectionLineParser.section_parser('link', '!virtual_sites4', context_type='link')
    @SectionLineParser.section_parser('link', '!virtual_sitesn', context_type='link')
    @SectionLineParser.section_parser('link', '!position_restraints', context_type='link')
    @SectionLineParser.section_parser('link', '!distance_restraints', context_type='link')
    @SectionLineParser.section_parser('link', '!dihedral_restraints', context_type='link')
    @SectionLineParser.section_parser('link', '!orientation_restraints', context_type='link')
    @SectionLineParser.section_parser('link', '!angle_restraints', context_type='link')
    @SectionLineParser.section_parser('link', '!angle_restraints_z', context_type='link')
    @SectionLineParser.section_parser('modification', 'bonds', context_type='modification')
    @SectionLineParser.section_parser('modification', 'angles', context_type='modification')
    @SectionLineParser.section_parser('modification', 'impropers', context_type='modification')
    @SectionLineParser.section_parser('modification', 'constraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'pairs', context_type='modification')
    @SectionLineParser.section_parser('modification', 'exclusions', context_type='modification')
    @SectionLineParser.section_parser('modification', 'pairs_nb', context_type='modification')
    @SectionLineParser.section_parser('modification', 'SETTLE', context_type='modification')
    @SectionLineParser.section_parser('modification', 'virtual_sites2', context_type='modification')
    @SectionLineParser.section_parser('modification', 'virtual_sites3', context_type='modification')
    @SectionLineParser.section_parser('modification', 'virtual_sites4', context_type='modification')
    @SectionLineParser.section_parser('modification', 'virtual_sitesn', context_type='modification')
    @SectionLineParser.section_parser('modification', 'position_restraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'distance_restraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'dihedral_restraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'orientation_restraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'angle_restraints', context_type='modification')
    @SectionLineParser.section_parser('modification', 'angle_restraints_z', context_type='modification')
    def _interactions(self, line, lineno=0, context_type=''):
        context = self.get_context(context_type)
        interaction_name = self.section[-1]
        delete = False
        if interaction_name.startswith('!'):
            interaction_name = interaction_name[1:]
            delete = True
        tokens = collections.deque(_tokenize(line))
        if tokens[0] == '#meta':
            _parse_meta(
                tokens,
                context,
                context_type=context_type,
                section=interaction_name,
            )
        else:
            n_atoms = self.interactions_natoms.get(interaction_name)
            _base_parser(
                tokens,
                context,
                context_type=context_type,
                section=interaction_name,
                natoms=n_atoms,
                delete=delete,
            )

    @SectionLineParser.section_parser('link', 'dihedrals', context_type='link')
    @SectionLineParser.section_parser('link', '!dihedrals', context_type='link')
    @SectionLineParser.section_parser('moleculetype', 'dihedrals', context_type='block')
    @SectionLineParser.section_parser('modification', 'dihedrals', context_type='modification')
    def _dih_interactions(self, line, lineno=0, context_type=''):
        context = self.get_context(context_type)
        interaction_name = self.section[-1]
        delete = False
        tokens = collections.deque(_tokenize(line))
        if tokens[0] == '#meta':
            _parse_meta(
                tokens,
                context,
                context_type=context_type,
                section=interaction_name,
            )
        else:
            n_atoms = self.interactions_natoms.get(interaction_name)
            _base_parser(
                tokens,
                context,
                context_type=context_type,
                section=interaction_name,
                natoms=n_atoms,
                delete=delete,
            )

          # Because of how they are described in gromacs, proper and improper
          # dihedral angles are all under the [ dihedrals ] section. However
          # the way they are treated differently in the library, at least on how
          # they generate edges. Here we move the all the impropers into their own
          # [ impropers ] section.

        propers = []
        impropers = context.interactions.get('impropers', [])
        for dihedral in context.interactions.get('dihedrals', []):
            if dihedral.parameters and dihedral.parameters[0] == '2':
                impropers.append(dihedral)
            else:
                propers.append(dihedral)

        context.interactions['dihedrals'] = propers
        context.interactions['impropers'] = impropers


    @SectionLineParser.section_parser('moleculetype', 'patterns')
    @SectionLineParser.section_parser('moleculetype', 'features')
    @SectionLineParser.section_parser('moleculetype', 'non-edge')
    @SectionLineParser.section_parser('modifications', 'non-edge')
    def _invalid_out_of_link(self, line, lineno=0):
        raise IOError('The "{}" section is only valid in links.'
                      .format(self.section[-1]))

    @SectionLineParser.section_parser('link', context_type='link')
    @SectionLineParser.section_parser('link', 'molmeta', context_type='molmeta')
    def _link(self, line, lineno=0, context_type=''):
        tokens = collections.deque(_tokenize(line))
        _parse_link_attribute(tokens, self.current_link, context_type)

    @SectionLineParser.section_parser('modification')
    def _modification(self, line, lineno=0):
        self.current_modification.name = line

    @SectionLineParser.section_parser('link', 'atoms')
    def _link_atoms(self, line, lineno=0):
        tokens = collections.deque(_tokenize(line))
        _parse_link_atom(tokens, self.current_link)

    @SectionLineParser.section_parser('modification', 'atoms')
    def _modification_atoms(self, line, lineno=0):
        tokens = collections.deque(_tokenize(line))
        _parse_link_atom(tokens, self.current_modification,
                         defaults={'PTM_atom': False},
                         treat_prefix=False)

    @SectionLineParser.section_parser('link', 'patterns', context_type='link')
    @SectionLineParser.section_parser('modification', 'patterns', context_type='modification')
    def _link_patterns(self, line, lineno=0, context_type=''):
        context = self.get_context(context_type)
        tokens = collections.deque(_tokenize(line))
        _parse_patterns(tokens, context, context_type)

    @SectionLineParser.section_parser('link', 'features', context_type='link')
    @SectionLineParser.section_parser('modification', 'features', context_type='modification')
    def _link_features(self, line, lineno=0, context_type=''):
        context = self.get_context(context_type)
        tokens = collections.deque(_tokenize(line))
        _parse_features(tokens, context, context_type)


def _some_atoms_left(tokens, atoms, natoms):
    """
    Return True if the token list expected to contain atoms.

    If the number of atoms is known before hand, then the function compares the
    number of already found atoms to the expected number. If the '--' token if
    found, it is removed from the token list and there is no atom left.

    Parameters
    ----------
    tokens: collections.deque[str]
        Deque of token to inspect. The deque **can be modified** in place.
    atoms: list
        List of already found atoms.
    natoms: int or None
        The number of expected atoms if known, else None.

    Returns
    -------
    bool
    """
    if not tokens:
        return False
    if tokens and tokens[0] == '--':
        tokens.popleft()
        return False
    if natoms is not None and len(atoms) >= natoms:
        return False
    return True


def _parse_atom_attributes(token):
    """
    Parse bracketed tokens.

    Parameters
    ----------
    token: str
        Token in the form of a json dictionary.

    Returns
    -------
    dict
    """
    if not token.strip().startswith('{'):
        raise ValueError('The token should start with a curly bracket.')
    try:
        attributes = json.loads(token)
    except JSONDecodeError as error:
        raise ValueError('The following value is not a valid atom attribute token: "{}".'
                         .format(token)) from error
    modifications = {}
    for key, value in attributes.items():
        try:
            if '|' in value:
                modifications[key] = Choice(value.split('|'))
        except TypeError:
            pass
    attributes.update(modifications)
    return attributes


def _get_atoms(tokens, natoms):
    atoms = []
    while tokens and _some_atoms_left(tokens, atoms, natoms):
        token = tokens.popleft()
        if token.startswith('{'):
            msg = 'Found atom attributes without an atom reference.'
            raise IOError(msg)
        if tokens:
            next_token = tokens[0]
        else:
            next_token = ''
        if next_token.startswith('{'):
            atoms.append([token, _parse_atom_attributes(next_token)])
            tokens.popleft()
        else:
            atoms.append([token, {}])
    return atoms


def _treat_block_interaction_atoms(atoms, context, section):
    atom_names = list(context.nodes)
    all_references = []
    for atom in atoms:
        reference = atom[0]
        if reference.isdigit():
            # The indices in the file are 1-based
            reference = int(reference) - 1
            try:
                reference = atom_names[reference]
            except IndexError:
                msg = ('There are {} atoms defined in the block "{}". '
                       'Interaction in section "{}" cannot refer to '
                       'atom index {}.')
                raise IOError(msg.format(len(context), context.name,
                                         section, reference + 1))
            atom[0] = reference
        else:
            if reference not in context:
                msg = ('There is no atom "{}" defined in the block "{}". '
                       'Section "{}" cannot refer to it.')
                raise IOError(msg.format(reference, context.name, section))
            if reference[0] in '+-<>':
                msg = ('Atom names in blocks cannot be prefixed with + or -. '
                       'The name "{}", used in section "{}" of the block "{}" '
                       'is not valid in a block.')
                raise IOError(msg.format(reference, section, context.name))
        all_references.append(reference)
    return all_references


def _split_node_key(key):
    """
    Split a node key into a prefix and a base and validate the key validity.
    """
    if not key:
        raise IOError('A node key cannot be empty.')

    # If the atom name is prefixed, we can get the order.
    prefix_end = 0  # Make sure prefix_end is defined even if key is empty
    for prefix_end, char in enumerate(key):
        if char not in '+-><*':
            break
    else:  # no break
        # This branch could also be taken if 'key' was empty. However, we
        # tested already that it was not the case.
        msg = ('The atom key "{}" is not valid. There must be a name '
               'following prefix.')
        raise IOError(msg.format(key))

    prefix = key[:prefix_end]
    if len(set(prefix)) > 1:
        msg = ('Atom name prefix cannot mix characters. Atom name "{}" '
               'is not a valid key.')
        raise IOError(msg.format(key))

    base = key[prefix_end:]

    return prefix, base


def _get_order_and_prefix_from_attributes(attributes):
    prefix_from_attributes = ''
    order_from_attributes = None
    Sequence = collections.abc.Sequence  # pylint: disable=invalid-name
    if attributes.get('order') is not None:
        order = attributes['order']
        order_from_attributes = order
        if isinstance(order, numbers.Integral) and not isinstance(order, bool):
            # Boolean as abstract subclasses of number.Integral as they can be
            # considered as 0 and 1. Yet, they yield unexpected results and
            # should not be accepted as valid values for 'order'.
            if order > 0:
                prefix_char = '+'
            else:
                prefix_char = '-'
            prefix_from_attributes = prefix_char * int(abs(order))
        elif (isinstance(order, Sequence)  # We can do the following operations
              and len(set(order)) == 1     # It is homogeneous
              and order[0] in '><*'):      # The character is an expected one
            prefix_from_attributes = order
        else:
            raise IOError('The order given in attribute ("{}") is not valid. '
                          'It must be am integer or a homogeneous series '
                          'of ">", "<", or "*".'
                          .format(order))
    return prefix_from_attributes, order_from_attributes


def _get_order_and_prefix_from_prefix(prefix):
    """
    Convert a prefix into a numerical value.
    """
    prefix_from_prefix = None
    order_from_prefix = 0
    if not prefix:
        return prefix_from_prefix, order_from_prefix

    # It is already validated.
    prefix_from_prefix = prefix
    if prefix[0] == '+':
        order_from_prefix = len(prefix)
    elif prefix[0] == '-':
        order_from_prefix = -len(prefix)
    else:
        order_from_prefix = prefix
    return prefix_from_prefix, order_from_prefix


def _treat_atom_prefix(reference, attributes):
    """
    Connect graph keys, order, and atom names.

    In a link, the graph keys, the order attribute of the atoms, and the atom
    names are interconnected. In most cases, the graph key is the atom name
    prefixed in a way that represent the order attribute. It is possible to
    define the order and the atom name from the graph key, or to set the graph
    key to represent the order, depending on what is explicitly specified in
    the file.

    In a link node, the order can be an integer, a series of '>' (*e.g.* '>',
    '>>', '>>>', ...), a series of '<', or a series of '*'. The series of '>',
    '<', and '*' translate directly from the key prefix to the order attribute,
    and *vice versa*. Numerical values of the order attribute, however, are
    converted into series of '+' or '-' for the key prefix; there, the number
    of '+' or '-' in the prefix corresponds to the value of the order
    attribute.

    The order can be specified either by the key prefix, or by the attribute.
    If it is specified in the two places, then they have to match each other or
    a :exc:`IOError` is raised.

    If the atom name is explicitly specified, then it is not modified. If it is
    not specified, then it is set from the node key. The base of the node key
    (*i.e.* what follows the prefix) is not modified, but a prefix can be added
    if there is none. The base of the node key and the atom name *can* differ.
    The atom name is what will be use for the graph isomorphism. The base of
    the key cannot be empty (*i.e.* '+++' or '*' are not valid keys); if it is,
    then an :exc:`IOError` is raised.

    Parameters
    ----------
    reference: str
        A node key for a link, as written in the file.
    attributes: dict
        The node attributes read fro the file.

    Returns
    -------
    prefixed_reference: str
        The node key with the appropriate prefix.
    attributes: dict
        A shalow copy of the node attribute dictionary with the 'order' and the
        'atomname' attributes set as appropriate.

    Raises
    ------
    IOError
        The node key, or an attribute value is invalid.

    Examples
    --------

    >>> _treat_atom_prefix('+BB', {})
    ('+BB', {'order': 1, 'atomname': 'BB'})
    >>> _treat_atom_prefix('BB', {'order': 1})
    ('+BB', {'order': 1, 'atomname': 'BB'})
    >>> _treat_atom_prefix('--XX', {'atomname': 'BB'})
    ('+BB', {'order': -2, 'atomname': 'BB'})
    >>> _treat_atom_prefix('>>BB', {})
    ('>>BB', {'order': '>>', 'atomname': 'BB'})

    """
    prefix, base = _split_node_key(reference)

    # Is the order specified in the attributes?
    (prefix_from_attributes,
     order_from_attributes) = _get_order_and_prefix_from_attributes(attributes)
    # Is there a specified prefix?
    (prefix_from_prefix,
     order_from_prefix) = _get_order_and_prefix_from_prefix(prefix)

    # If the order is defined twice, is it consistent?
    if (order_from_attributes is not None
            and prefix_from_prefix is not None
            and order_from_attributes != order_from_prefix):
        msg = ('The sequence order for atom "{}" of a '
               'link is not consistent between the name prefix '
               '(order={}) and the atom attributes (order={}).')
        raise IOError(msg.format(reference, order_from_prefix, order_from_attributes))

    return_attributes = copy.copy(attributes)
    if order_from_attributes is None:
        return_attributes['order'] = order_from_prefix
    if 'atomname' not in return_attributes:
        return_attributes['atomname'] = base
    if prefix_from_prefix is None:
        prefixed = prefix_from_attributes + base
    else:
        prefixed = reference

    return prefixed, return_attributes


def _treat_link_interaction_atoms(atoms, context, section):
    all_references = []
    for reference, attributes in atoms:
        intermediate = context._apply_to_all_nodes.copy()
        intermediate.update(attributes)
        attributes = intermediate

        prefixed_reference, attributes = _treat_atom_prefix(reference, attributes)
        all_references.append(prefixed_reference)

        if prefixed_reference in context:
            context_atom = context.nodes[prefixed_reference]
            for key, value in attributes.items():
                if key in context_atom and value != context_atom[key]:
                    msg = ('Attribute {} of atom {} conflicts in a link '
                           'between its definition in section "{}" '
                           '(value is "{}") and its previous definition '
                           '(value was "{}").')
                    raise IOError(msg.format(key, reference, section,
                                             value, context_atom[key]))
            context_atom.update(attributes)
        else:
            context.add_node(prefixed_reference, **attributes)
    return all_references


def _parse_interaction_parameters(tokens):
    parameters = []
    for token in tokens:
        if _is_param_effector(token):
            effector_name, effector_param_str = token.split('(', 1)
            effector_param_str = effector_param_str[:-1]  # Remove the closing parenthesis
            try:
                effector_class = PARAMETER_EFFECTORS[effector_name]
            except KeyError:
                raise IOError('{} is not a known parameter effector.'
                              .format(effector_name))
            if '|' in effector_param_str:
                effector_param_str, effector_format = effector_param_str.split('|')
            else:
                effector_format = None
            effector_param = [elem.strip() for elem in effector_param_str.split(',')]
            parameter = effector_class(effector_param, format_spec=effector_format)
        else:
            parameter = token
        parameters.append(parameter)
    return parameters


def _is_param_effector(token):
    return (
        '(' in token
        and not token.startswith('(')
        and token.endswith(')')
    )


def _base_parser(tokens, context, context_type, section, natoms=None, delete=False):
    if context_type != 'link' and delete:
        raise IOError('Interactions can only be removed in links.')
    delimiter_count = tokens.count('--')
    if delimiter_count > 1:
        msg = 'There can be 0 or 1 "--" delimiter; {} found.'
        raise IOError(msg.format(delimiter_count))

    # Group the atoms and their attributes
    atoms = _get_atoms(tokens, natoms)
    if natoms is not None and len(atoms) != natoms:
        raise IOError('Found {} atoms while {} were expected.'
                      .format(len(atoms), natoms))

    # Normalize the atom references.
    # Blocks and links treat these references differently.
    # For blocks:
    # * references can be written as indices or atom names
    # * a reference cannot be prefixed by + or -
    # * an interaction cannot create a new atom
    # For links:
    # * references must be atom names, but they can be prefixed with one or
    #   more + or - to signify the order in the sequence
    # * interactions create nodes
    if context_type == 'block':
        treated_atoms = _treat_block_interaction_atoms(atoms, context, section)
    elif context_type in ('link', 'modifications'):
        treated_atoms = _treat_link_interaction_atoms(atoms, context, section)


    # Getting the atoms consumed the "--" delimiter if any. So what is left
    # are the interaction parameters or the meta attributes.
    if tokens and tokens[-1].startswith('{'):
        token = tokens.pop()
        meta = json.loads(token)
    else:
        meta = {}
    parameters = _parse_interaction_parameters(tokens)

    apply_to_all_interactions = context._apply_to_all_interactions[section]
    meta = dict(collections.ChainMap(meta, apply_to_all_interactions))

    if delete:
        interaction = DeleteInteraction(
            atoms=treated_atoms,
            atom_attrs=[atom[1] for atom in atoms],
            parameters=parameters,
            meta=meta,
        )
        interaction_list = context.removed_interactions.get(section, [])
        interaction_list.append(interaction)
        context.removed_interactions[section] = interaction_list
    else:
        interaction = Interaction(
            atoms=treated_atoms,
            parameters=parameters,
            meta=meta,
        )
        interaction_list = context.interactions.get(section, [])
        interaction_list.append(interaction)
        context.interactions[section] = interaction_list


def _parse_block_atom(tokens, context):
    if tokens[-1].startswith('{'):
        attributes = _parse_atom_attributes(tokens.pop())
    else:
        attributes = {}

    # deque does not support slicing
    first_six = (tokens.popleft() for _ in range(6))
    _, atype, resid, resname, name, charge_group = first_six
    if name in context:
        msg = ('There is already an atom named "{}" in the block "{}". '
               'Atom names must be unique within a block.')
        raise IOError(msg.format(name, context.name))
    atom = {
        'atomname': name,
        'atype': atype,
        'resname': resname,
        'resid': int(resid),
        'charge_group': int(charge_group),
    }
    # charge and mass are optional, but charge has to be defined for mass to be
    if tokens:
        atom['charge'] = float(tokens.popleft())
    if tokens:
        atom['mass'] = float(tokens.popleft())
    context.add_atom(dict(collections.ChainMap(attributes, atom)))


def _parse_link_atom(tokens, context, defaults=None, treat_prefix=True):
    if len(tokens) > 2:
        raise IOError('Unexpected column in link atom definition.')
    if len(tokens) < 2:
        raise IOError('Missing column in link atom definition.')
    reference = tokens[0]
    if defaults is None:
        defaults = {}
    attributes = _parse_atom_attributes(tokens[1])
    if treat_prefix:
        prefixed_reference, attributes = _treat_atom_prefix(reference, attributes)
    else:
        prefixed_reference = reference
        attributes['atomname'] = reference
    attributes = dict(collections.ChainMap(attributes, context._apply_to_all_nodes))
    node_attributes = context.nodes.get(prefixed_reference, {})
    for attr, value in attributes.items():
        if value != node_attributes.get(attr, value):
            msg = ('Conflict in an atom attributes in the definition of a '
                   'link node. Cannot set the attribute "{}" of atom "{}" '
                   'to "{}" because it is already defined as "{}".')
            raise IOError(msg.format(attr, reference, value,
                                     node_attributes[node_attributes[attr]]))
    full_attributes = dict(collections.ChainMap(attributes, node_attributes, defaults))

    if prefixed_reference in context.nodes:
        context.nodes[prefixed_reference] = full_attributes
    else:
        context.add_node(prefixed_reference, **full_attributes)


def _parse_link_attribute(tokens, context, section):
    if len(tokens) > 2:
        raise IOError('Unexpected column in section "{}".'.format(section))
    if len(tokens) < 2:
        raise IOError('Missing column in section "{}".'.format(section))
    key, value = tokens
    if '|' in value:
        value = Choice(json.loads(value).split('|'))
    elif '(' in value and value.endswith(')') and not value.startswith('('):
        open_pos = value.find('(')
        function = value[:open_pos]
        argument = json.loads(value[open_pos + 1:-1])
        value = VALUE_PREDICATES[function](argument)
    else:
        value = json.loads(value)

    if section == 'link':
        context._apply_to_all_nodes[key] = value
    elif section == 'molmeta':
        context.molecule_meta[key] = value
    else:
        raise ValueError('Parser only defined for sections "link" and "molmeta".')


def _parse_meta(tokens, context, context_type, section):
    """
    Parse lines starting with '#meta'.

    The function expects 2 tokens. The first token is assumed to be '#meta' and
    is ignored. The second must be a bracketed token. This second token is
    parsed as a dictionary and updated the dictionary of attributes to add to
    all the nodes involved in a given type of interaction for the context.

    The type of interaction is set by the `section` argument.

    The context is a :class:`vermouth.molecule.Block` or a subclass such as a
    :class:`vermouth.molecule.Link`.

    The `context_type` is a string version of the first level section (i.e
    "link", "block", or "modification").
    """
    if len(tokens) > 2:
        msg = ('Unexpected column when defining meta attributes for section '
               '"{}" of a {}. {} tokens read instead of 2.')
        raise IOError(msg.format(section, context_type, len(tokens)))
    if len(tokens) < 2:
        msg = 'Missing column when defining meta attributes for section "{}" of a {}.'
        raise IOError(msg.format(section, context_type))
    attributes = json.loads(tokens[-1])
    context._apply_to_all_interactions[section].update(attributes)


def _parse_edges(tokens, context, context_type, negate):
    if negate and context_type != 'link':
        raise IOError('The "non-edges" section is only valid in links.')
    atoms = _get_atoms(tokens, natoms=2)
    prefixed_atoms = []
    for atom in atoms:
        prefixed_reference, attributes = _treat_atom_prefix(*atom)
        try:
            apply_to_all_nodes = context._apply_to_all_nodes
        except AttributeError:
            apply_to_all_nodes = {}
        full_attributes = dict(collections.ChainMap(attributes, apply_to_all_nodes))
        prefixed_atoms.append([prefixed_reference, full_attributes])
    if negate:
        context.non_edges.append([prefixed_atoms[0][0], prefixed_atoms[1][1]])
    else:
        error_message = 'Atom with name {} not found for {} {}'
        for prefixed_atom in prefixed_atoms:
            atomname = prefixed_atom[0]
            if atomname not in context and context_type == 'modification':
                raise KeyError(error_message.format(atomname, context_type,
                                                    context.name))
        context.add_edge(prefixed_atoms[0][0], prefixed_atoms[1][0])


def _parse_patterns(tokens, context, context_type):
    if context_type != 'link':
        raise IOError('The "partterns" section is only valid in links.')
    atoms = _get_atoms(tokens, natoms=None)
    context.patterns.append(atoms)


def _parse_variables(tokens, force_field, section):
    if len(tokens) > 2:
        raise IOError('Unexpected column in section "{}".'.format(section))
    if len(tokens) < 2:
        raise IOError('Missing column in section "{}".'.format(section))
    key, value = tokens
    try:
        value = json.loads(value)
    except JSONDecodeError:
        value = str(value)
    force_field.variables[key] = value


def _parse_features(tokens, context, context_type):
    if context_type != 'link':
        raise IOError('The "features" section is only valid in links.')
    context.features.update(set(tokens))


def read_ff(lines, force_field):
    director = FFDirector(force_field)
    return list(director.parse(iter(lines)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the backmapping file parser and its utilities.
"""

# We call "private" methods and functions as part of the tests, which pylint
# does not like. Because we want to test these methods and functions, we have
# to call them outside of there normal "scope".
# pylint: disable=protected-access

# Pylint wrongly complains about the use of pytest fixtures.
# pylint: disable=redefined-outer-name

import collections
import itertools
import textwrap
from pathlib import Path

import pytest

import vermouth
import vermouth.map_input
import vermouth.forcefield

Reference = collections.namedtuple('Reference',
                                   'string name from_ff to_ff mapping weights extra')

# Test system 1: simple case
# from: a-b-c-d-e-f
#       ----- -----
# to:     A     B
SYSTEM_BASIC = Reference(
    string="""\
system1

[atoms]
1 a A
2 b A
3 c A
4 d B
5 e B
6 f B
""",
    name='system1',
    from_ff=['universal'],  # default
    to_ff=['martini22'],  # default
    mapping={
        (0, 'a'): [(0, 'A')],
        (0, 'b'): [(0, 'A')],
        (0, 'c'): [(0, 'A')],
        (0, 'd'): [(0, 'B')],
        (0, 'e'): [(0, 'B')],
        (0, 'f'): [(0, 'B')],
    },
    weights={
        (0, 'A'): {(0, 'a'): 1, (0, 'b'): 1, (0, 'c'): 1},
        (0, 'B'): {(0, 'd'): 1, (0, 'e'): 1, (0, 'f'): 1},
    },
    extra=[],
)

# Test system 1: shared atoms
# from: a-b-c-d-e-f
#       -------
#           -------
# to:     A     B
# c has a greater weight in A, while d has a greater weight in B
SYSTEM_SHARED = Reference(
    string="""\
system2

[atoms]
1 a A
2 b A
3 c A B A  ; more on A (also comment)
4 d A B B  ; more on B
5 e B
6 f B
""",
    name='system2',
    from_ff=['universal'],  # default
    to_ff=['martini22'],  # default
    mapping={
        (0, 'a'): [(0, 'A')],
        (0, 'b'): [(0, 'A')],
        (0, 'c'): [(0, 'A'), (0, 'B'), (0, 'A')],
        (0, 'd'): [(0, 'A'), (0, 'B'), (0, 'B')],
        (0, 'e'): [(0, 'B')],
        (0, 'f'): [(0, 'B')],
    },
    weights={
        (0, 'A'): {(0, 'a'): 1, (0, 'b'): 1, (0, 'c'): 2 / 3, (0, 'd'): 1 / 3},
        (0, 'B'): {(0, 'c'): 1 / 3, (0, 'd'): 2 / 3, (0, 'e'): 1, (0, 'f'): 1},
    },
    extra=[],
)

# Test system 3: unmapped particles
# Same as system 1, but C, D, and E do not have underlying atoms
SYSTEM_EXTRA = Reference(
    string=SYSTEM_BASIC.string + """
[extra]
C ; alone on its line
D E  ; two particles on the same line
""",
    name=SYSTEM_BASIC.name,
    from_ff=SYSTEM_BASIC.from_ff,
    to_ff=SYSTEM_BASIC.to_ff,
    mapping=SYSTEM_BASIC.mapping,
    weights=SYSTEM_BASIC.weights,
    extra=['C', 'D', 'E'],
)

# Test system 4: null-weights
# from: a-b-c-d-e-f
#       -------
#           -------
# to:     A     B
# c as a weight of 0 on B, and d has a weight of 0 on A
SYSTEM_NULL = Reference(
    string="""\
system4

[atoms]
1 a A
2 b A
3 c A !B
4 d !A B
5 e B
6 f B
""",
    name='system4',
    from_ff=['universal'],  # default
    to_ff=['martini22'],  # default
    mapping={
        (0, 'a'): [(0, 'A')],
        (0, 'b'): [(0, 'A')],
        (0, 'c'): [(0, 'A'), (0, 'B')],
        (0, 'd'): [(0, 'A'), (0, 'B')],
        (0, 'e'): [(0, 'B')],
        (0, 'f'): [(0, 'B')],
    },
    weights={
        (0, 'A'): {(0, 'a'): 1, (0, 'b'): 1, (0, 'c'): 1, (0, 'd'): 0},
        (0, 'B'): {(0, 'c'): 0, (0, 'd'): 1, (0, 'e'): 1, (0, 'f'): 1},
    },
    extra=[],
)

# Test system 5: origin and target force fields
# Same as system 1, but with origin and target force field explicitly
# specified.
SYSTEM_FROM_TO = Reference(
    string=SYSTEM_BASIC.string + """

[from]
from_ff
from_ff_2
from_ff_3 from_ff_4

[mapping]
from_ff_5 from_ff_6
from_ff_7

[to]
to_ff
to_ff_2 to_ff_3

""",
    name=SYSTEM_BASIC.name,
    from_ff=['from_ff', 'from_ff_2', 'from_ff_3',
             'from_ff_4', 'from_ff_5', 'from_ff_6', 'from_ff_7'],
    to_ff=['to_ff', 'to_ff_2', 'to_ff_3'],
    mapping=SYSTEM_BASIC.mapping,
    weights=SYSTEM_BASIC.weights,
    extra=SYSTEM_BASIC.extra,
)

# Test system 6: unused field from Backward
# Same as system 1, but with a section that has to be ignored.
SYSTEM_IGNORE = Reference(
    string=SYSTEM_BASIC.string + """
[ chiral ]
A B C D
""",
    name=SYSTEM_BASIC.name,
    from_ff=SYSTEM_BASIC.from_ff,
    to_ff=SYSTEM_BASIC.to_ff,
    mapping=SYSTEM_BASIC.mapping,
    weights=SYSTEM_BASIC.weights,
    extra=SYSTEM_BASIC.extra,
)


def case_to_dummy_ffs(ff_names, names, mapping, weights, extra):
    from_atoms = []
    for _, atname in mapping.keys():
        from_atoms.append(atname)

    to_atoms = []
    for _, atname in weights.keys():
        to_atoms.append(atname)

    ffs = {}
    for ff_name in ff_names:
        ff = vermouth.forcefield.ForceField(name=ff_name)
        ffs[ff_name] = ff
        for name in names:
            block = vermouth.molecule.Block(force_field=ff, name=name)
            block.add_nodes_from(from_atoms + to_atoms)
            for node in block.nodes:
                block.nodes[node]['atomname'] = node
            ff.blocks[name] = block
    return ffs


def mapping_to_old_style(mapping):
    map_out = collections.defaultdict(list)
    weights = collections.defaultdict(dict)
    for node_from in mapping:
        for node_to, weight in mapping[node_from].items():
            map_out[(0, node_from)].append((0, node_to))
            weights[(0, node_to)][(0, node_from)] = weight
    return dict(map_out), dict(weights)


def compare_old_new_mappings(mappings, reference):
    assert mappings.keys() == reference.keys()
    for from_ff in mappings:
        assert mappings[from_ff].keys() == reference[from_ff].keys()
        for to_ff in mappings[from_ff]:
            assert mappings[from_ff][to_ff].keys() == reference[from_ff][to_ff].keys()
            for case_name in mappings[from_ff][to_ff]:
                mapping, weights = mapping_to_old_style(mappings[from_ff][to_ff][case_name].mapping)
                assert mapping.keys() == reference[from_ff][to_ff][case_name][0].keys()
                for from_, tos in mapping.items():
                    # Old style mappigns could contain something like:
                    #  {(0, 'd'): [(0, 'A'), (0, 'B'), (0, 'B')]}. New style
                    # mappings can't. Instead, it's in the weights.
                    assert all(to in reference[from_ff][to_ff][case_name][0][from_] for to in tos)
                assert weights == reference[from_ff][to_ff][case_name][1]
                assert mappings[from_ff][to_ff][case_name].block_to.extra == reference[from_ff][to_ff][case_name][2]


@pytest.mark.parametrize(
    'case',
    [SYSTEM_BASIC, SYSTEM_SHARED, SYSTEM_EXTRA, SYSTEM_NULL, SYSTEM_FROM_TO, SYSTEM_IGNORE]
)
def test_read_mapping_partial(case):
    """
    Test that regular mapping files are read as expected.
    """
    full_mapping = vermouth.map_input._read_mapping_partial(case.string.split('\n'), 1)
    name, from_ff, to_ff, mapping, weights, extra, _ = full_mapping
    assert name == case.name
    assert from_ff == case.from_ff
    assert to_ff == case.to_ff
    assert mapping == case.mapping
    assert extra == case.extra
    assert weights == case.weights


@pytest.mark.parametrize('content', (
    """
dummy

[ atoms ]
0 X1 !A A B
    """,  # Inconsistent bead weight
    """
[ molecule
dummy
    """,  # Incomplete section line
    """
dummy

[ atoms ]
0 X1 A B
1 X2 C D
2 X1 Y U
    """,  # Multiple difinitions for the same atom
    """
[ atoms ]
0 A B
    """,  # no molecule name
    """
Just a pile of garbage.
Clearly, this is not a mapping file.
Not even a partial one.
    """,
    # If a file contains two consecutive [ molecule ] lines, it means the first
    # of the molecules is empty.
    """
[ molecule ]
    """,
))
def test_read_mapping_errors(content):
    """
    Test that syntax error are caught when reading a mapping.
    """
    with pytest.raises(IOError):
        vermouth.map_input._read_mapping_partial(content.split('\n'), 1)


@pytest.mark.parametrize(
    'case',
    [SYSTEM_BASIC, SYSTEM_SHARED, SYSTEM_EXTRA, SYSTEM_NULL, SYSTEM_FROM_TO, SYSTEM_IGNORE]
)
def test_read_mapping_file(case):
    """
    Test that regular mapping files are read as expected.
    """
    reference = collections.defaultdict(lambda: collections.defaultdict(dict))
    for from_ff, to_ff in itertools.product(case.from_ff, case.to_ff):
        reference[from_ff][to_ff][case.name] = (
            case.mapping, case.weights, case.extra
        )

    ffs = case_to_dummy_ffs(case.from_ff + case.to_ff, [case.name], case.mapping,
                            case.weights, case.extra)

    reference = vermouth.map_input._default_to_dict(reference)

    mappings = vermouth.map_input.read_backmapping_file(
        ['[ molecule ]'] + case.string.split('\n'),
        ffs
    )
    compare_old_new_mappings(mappings, reference)


@pytest.fixture
def reference_multi():
    """
    Build a reference file and mapping collection with multiple molecules.
    """
    content = textwrap.dedent("""
        ; Some line before the first molecule.
        ; Just because.

        [ molecule ]
        dummy_0

        [ atoms ]
        0 X1 A
        1 X2 B

        ; Some mess between two molecules.

        [ molecule ]
        ; A comment just after the molecule section.
        dummy_1

        [ atoms ]
        0 X2 C
        1 X3 D
    """).split('\n')
    reference = {'universal': {'martini22': {
        'dummy_0': (
            {(0, 'X1'): [(0, 'A')], (0, 'X2'): [(0, 'B')]},
            {(0, 'A'): {(0, 'X1'): 1.0}, (0, 'B'): {(0, 'X2'): 1.0}},
            [],
        ),
        'dummy_1': (
            {(0, 'X2'): [(0, 'C')], (0, 'X3'): [(0, 'D')]},
            {(0, 'C'): {(0, 'X2'): 1.0}, (0, 'D'): {(0, 'X3'): 1.0}},
            [],
        ),
    }}}
    return content, reference


def test_read_mapping_file_multiple(reference_multi):
    """
    Test that read_mapping_file can read more than one molecule.
    """
    content, reference = reference_multi
    from_names = list(reference.keys())
    to_names = []
    block_names = []

    for k in reference:
        to_names.extend(reference[k].keys())
        for to in reference[k]:
            block_names.extend(reference[k][to].keys())
    force_fields = case_to_dummy_ffs(from_names + to_names, block_names,
                                     {(0, 'X1'): [(0, 'A')], (0, 'X2'): [(0, 'B')], (0, 'X3'): [(0, 'D')]},
                                     {(0, 'A'): {(0, 'X1'): 1.0}, (0, 'B'): {(0, 'X2'): 1.0}, (0, 'C'): {(0, 'X2'): 1.0}, (0, 'D'): {(0, 'X3'): 1.0}},
                                     [])
    mappings = vermouth.map_input.read_backmapping_file(content, force_fields)
    compare_old_new_mappings(mappings, reference)


@pytest.fixture(scope='session')
def ref_mapping_directory(tmpdir_factory):
    """
    Build a file tree with mapping files.
    """
    basedir = tmpdir_factory.mktemp('data')
    mapdir = basedir.mkdir('mappings')

    template = textwrap.dedent("""
        [ molecule ]
        dummy_{0}

        [ from ]
        {1}

        [ to ]
        {2}

        [ atoms ]
        0 X1{0} A{0} B{0}
        1 X2{0} C{0} D{0}
    """)

    mappings = collections.defaultdict(lambda: collections.defaultdict(dict))

    force_fields_from = ['ff{}'.format(i) for i in range(4)]
    force_fields_to = force_fields_from + ['only_to']
    force_fields_from = force_fields_from + ['only_from']
    iterate_on = itertools.product(force_fields_from, force_fields_to, range(3))
    for idx, (from_ff, to_ff, _) in enumerate(iterate_on):
        mapfile = mapdir / 'file{}.map'.format(idx)
        with open(str(mapfile), 'w') as outfile:
            outfile.write(template.format(idx, from_ff, to_ff))

        mapping = {
            (0, 'X1{}'.format(idx)): [(0, 'A{}'.format(idx)), (0, 'B{}'.format(idx))],
            (0, 'X2{}'.format(idx)): [(0, 'C{}'.format(idx)), (0, 'D{}'.format(idx))],
        }
        weights = {
            (0, 'A{}'.format(idx)): {(0, 'X1{}'.format(idx)): 0.5},
            (0, 'B{}'.format(idx)): {(0, 'X1{}'.format(idx)): 0.5},
            (0, 'C{}'.format(idx)): {(0, 'X2{}'.format(idx)): 0.5},
            (0, 'D{}'.format(idx)): {(0, 'X2{}'.format(idx)): 0.5},
        }
        extra = []
        mappings[from_ff][to_ff]['dummy_{}'.format(idx)] = (mapping, weights, extra)

    mappings = {from_ff: dict(to_ff) for from_ff, to_ff in mappings.items()}

    return Path(str(basedir)), mappings


def test_read_mapping_directory(ref_mapping_directory):
    """
    Test that mapping files from a directory are propely found and read.
    """
    dirpath, ref_mappings = ref_mapping_directory
    from_names = list(ref_mappings.keys())
    to_names = []
    block_names = []
    mapping = {}
    weights = {}


    for k in ref_mappings:
        to_names.extend(ref_mappings[k].keys())
        for to in ref_mappings[k]:
            block_names.extend(ref_mappings[k][to].keys())
            for block_name in ref_mappings[k][to]:
                m, w, _ = ref_mappings[k][to][block_name]
                mapping.update(m)
                weights.update(w)
    force_fields = case_to_dummy_ffs(from_names + to_names, block_names,
                                     mapping, weights, [])


    mappings = vermouth.map_input.read_mapping_directory(dirpath, force_fields)
    compare_old_new_mappings(mappings, ref_mappings)
#    assert mappings == ref_mappings


def test_read_mapping_directory_not_dir():
    """
    Test that :func:`vermouth.map_input.read_mapping_directory` fails when
    the input is not a directory.
    """
    with pytest.raises(NotADirectoryError):
        vermouth.map_input.read_mapping_directory('not a directory', {})


def test_read_mapping_directory_error(tmpdir):
    """
    Test that :func:`vermouth.map_input.read_mapping_directory` raises an
    exception when a file could not be read.
    """
    mapdir = Path(str(tmpdir.mkdir('mappings')))
    with open(str(mapdir / 'valid.backmap'), 'w') as outfile:
        outfile.write(textwrap.dedent("""
            [ molecule ]
            valid
            [ atoms ]
            0 A B
        """))
    with open(str(mapdir / 'not_valid.map'), 'w') as outfile:
        outfile.write('invalid content')
    with pytest.raises(IOError):
        vermouth.map_input.read_mapping_directory(mapdir, {})


def test_generate_self_mapping():
    """
    Test that :func:`vermouth.map_input.generate_self_mappings` works as
    expected.
    """
    # Build the input blocks
    blocks = {
        'A0': vermouth.molecule.Block([['AA', 'BBB'], ['BBB', 'CCCC']]),
        'B1': vermouth.molecule.Block([['BBB', 'CCCC'], ['BBB', 'E']]),
    }
    for name, block in blocks.items():
        block.name = name
        for atomname, node in block.nodes.items():
            node['atomname'] = atomname
    # Build the expected output
    ref_mappings = {
        'A0': (
            # mapping
            {"AA": {"AA": 1}, "BBB": {"BBB": 1}, "CCCC": {"CCCC": 1}},
            # extra
            [],
        ),
        'B1': (
            # mapping
            {"BBB": {"BBB": 1}, "CCCC": {"CCCC": 1}, "E": {"E": 1}},
            # extra
            [],
        ),
    }
    # Actually test
    mappings = vermouth.map_input.generate_self_mappings(blocks)
    assert mappings.keys() == ref_mappings.keys()
    for blockname in mappings:
        assert mappings[blockname].mapping, mappings[blockname].extras == ref_mappings[blockname]


def test_generate_all_self_mappings():
    """
    Test that :func:`vermouth.map_input.generate_all_self_mappings` generate
    the expected entries.
    """
    force_fields = []
    expected = []
    for idx in range(3):
        idx_str = str(idx)
        ff_name = 'ff_' + idx_str
        force_field = vermouth.forcefield.ForceField(name=ff_name)
        force_field.blocks = {
            'A' + idx_str: vermouth.molecule.Block([['AA', 'BBB'], ['BBB', 'CCCC']]),
            'B' + idx_str: vermouth.molecule.Block([['BBB', 'CCCC'], ['BBB', 'E']]),
        }
        for name, block in force_field.blocks.items():
            block.name = name
            for atomname, node in block.nodes.items():
                node['atomname'] = atomname
        force_fields.append(force_field)

        expected.append((ff_name, (ff_name, )))

    mappings = vermouth.map_input.generate_all_self_mappings(force_fields)
    found = [(from_ff, tuple(to_ff.keys())) for from_ff, to_ff in mappings.items()]

    # In python <= 3.5, dicts are not ordered.
    assert set(found) == set(expected)


@pytest.fixture
def base_mappings():
    """
    Build a basic (empty) mapping collection to modify.
    """
    return {
        'from_1': {
            'to_1': {
                'mol_1': ({}, {}, []),
                'mol_2': ({}, {}, []),
            },
        },
    }


@pytest.mark.parametrize('partial_mappings, expected', (
    (  # new force field from and to
        # partial
        {'from_2': {'to_2': {'mol_1': ({}, {}, [])}}},
        # expected
        {
            'from_1': {'to_1': {'mol_1': ({}, {}, []), 'mol_2': ({}, {}, [])}},
            'from_2': {'to_2': {'mol_1': ({}, {}, [])}}
        }
    ),
    (  # Add a target to an existing force field
        # partial
        {'from_1': {'to_2': {'mol_1': ({}, {}, [])}}},
        # expected
        {
            'from_1': {
                'to_1': {'mol_1': ({}, {}, []), 'mol_2': ({}, {}, [])},
                'to_2': {'mol_1': ({}, {}, [])}
            }
        }
    ),
    (  # Add a molecule to an existing force field pair
        # partial
        {'from_1': {'to_1': {'mol_3': ({}, {}, [])}}},
        # expected
        {'from_1': {'to_1': {
            'mol_1': ({}, {}, []),
            'mol_2': ({}, {}, []),
            'mol_3': ({}, {}, []),
        }}}
    ),
    (  # Replace a molecule from an existing force field pair
        # partial
        {'from_1': {'to_1': {'mol_1': ({}, {}, ['modified'])}}},
        # expected
        {'from_1': {'to_1': {
            'mol_1': ({}, {}, ['modified']),
            'mol_2': ({}, {}, []),
        }}}
    ),
))
def test_combine_mappings(base_mappings, partial_mappings, expected):
    """
    Test that :func:`vermouth.map_input.combine_mappings` works as expected.
    """
    vermouth.map_input.combine_mappings(base_mappings, partial_mappings)
    assert base_mappings == expected



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test PTM detection and canonicalisation.
"""

import networkx as nx
import pytest

import vermouth
import vermouth.processors.canonicalize_modifications as canmod

# pylint: disable=redefined-outer-name


@pytest.fixture
def known_ptm_graphs():
    """
    Provide some known PTMs in a defined order
    """
    ptm_graphs = []

    nh_ptm = nx.Graph(name='NH')
    nh_ptm.add_nodes_from([
        (0, {'atomname': 'H', 'PTM_atom': True, 'element': 'H'}),
        (1, {'atomname': 'N', 'PTM_atom': False, 'element': 'N'}),
    ])
    nh_ptm.add_edge(0, 1)
    ptm_graphs.append(nh_ptm)

    cooc_ptm = nx.Graph(name='COOC')
    cooc_ptm.add_nodes_from([
        (0, {'atomname': 'C', 'PTM_atom': False, 'element': 'C'}),
        (1, {'atomname': 'O', 'PTM_atom': True, 'element': 'O'}),
        (2, {'atomname': 'O', 'PTM_atom': True, 'element': 'O'}),
        (3, {'atomname': 'C', 'PTM_atom': False, 'element': 'C'}),
    ])
    cooc_ptm.add_edges_from([(0, 1), (1, 2), (2, 3)])
    ptm_graphs.append(cooc_ptm)

    return sorted(ptm_graphs, key=len, reverse=True)


def make_molecule(atoms, edges):
    """
    Makes molecules from atoms and edges
    """
    mol = vermouth.molecule.Molecule()
    mol.add_nodes_from(atoms.items())
    mol.add_edges_from(edges)
    return mol


@pytest.mark.parametrize('atoms, edges, expected', [
    ({}, [], []),
    (
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': True, 'element': 'N', 'resid': 1},
        },
        [(0, 1), (1, 2)],
        [({1, 2}, {0})]
    ),
    (
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': False, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': True, 'element': 'N', 'resid': 1},
        },
        [(0, 1), (1, 2)],
        [({2}, {1})]
    ),
    (
        {
            0: {'atomname': 'N', 'PTM_atom': True, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': False, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': True, 'element': 'N', 'resid': 1},
        },
        [(0, 1), (1, 2)],
        [({0}, {1}), ({2}, {1})]
    ),
    (
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 2},
        },
        [(0, 1), (1, 2)],
        [({1}, {0, 2})]
    ),
    (
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': True, 'element': 'N', 'resid': 2},
            3: {'atomname': 'CA', 'PTM_atom': False, 'element': 'N', 'resid': 2},
        },
        [(0, 1), (1, 2), (2, 3)],
        [({1, 2}, {0, 3})]
    ),
    (
        {
            0: {'atomname': 'A', 'PTM_atom': True, 'element': 'A', 'resid': 1},
            1: {'atomname': 'A', 'PTM_atom': True, 'element': 'A', 'resid': 1},
            2: {'atomname': 'A', 'PTM_atom': False, 'element': 'A', 'resid': 1},
            3: {'atomname': 'A', 'PTM_atom': False, 'element': 'A', 'resid': 1},
            4: {'atomname': 'A', 'PTM_atom': False, 'element': 'A', 'resid': 2},
            5: {'atomname': 'A', 'PTM_atom': False, 'element': 'A', 'resid': 1},
            6: {'atomname': 'A', 'PTM_atom': False, 'element': 'A', 'resid': 3},
        },
        [(0, 1), (0, 2), (2, 3), (3, 4), (2, 5), (2, 6)],
        [({0, 1}, {2})]
    )
])
def test_ptm_groups(atoms, edges, expected):
    """
    Make sure PTM atoms are grouped correctly with appropriate anchors
    """
    molecule = make_molecule(atoms, edges)

    found = canmod.find_ptm_atoms(molecule)
    assert expected == found


@pytest.mark.parametrize('atoms, edges, expected', [
    pytest.param(
        # This needs to raise a KeyError, because not all the anchors are
        # covered. This is the root of #140
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 2},
        },
        [(0, 1), (1, 2)],
        [('NH', {1: 0, 2: 1})],
        marks=pytest.mark.xfail(raises=KeyError, strict=True)
    ),
    (
        # Simplest case: one PTM atom for 1 residue
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
        },
        [(0, 1)],
        [('NH', {0: 1, 1: 0})]
    ),
    (
        # Two PTM atoms with a shared anchor
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
        },
        [(0, 1), (0, 2)],
        [('NH', {0: 1, 1: 0}), ('NH', {0: 1, 2: 0})]
    ),
    (
        # Two PTM atoms with two anchors covered (?) by 2 fragments
        {
            0: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 1},
            1: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 1},
            2: {'atomname': 'H', 'PTM_atom': True, 'element': 'H', 'resid': 2},
            3: {'atomname': 'N', 'PTM_atom': False, 'element': 'N', 'resid': 2},
        },
        [(0, 1), (1, 2), (2, 3)],
        [('NH', {0: 1, 1: 0}), ('NH', {2: 0, 3: 1})]
    ),
    (
        # Two PTM atoms with two anchors covered by 1 fragment
        {
            0: {'atomname': 'C', 'PTM_atom': False, 'element': 'C', 'resid': 1},
            1: {'atomname': 'O', 'PTM_atom': True, 'element': 'O', 'resid': 1},
            2: {'atomname': 'O', 'PTM_atom': True, 'element': 'O', 'resid': 2},
            3: {'atomname': 'C', 'PTM_atom': False, 'element': 'C', 'resid': 2},
        },
        [(0, 1), (1, 2), (2, 3)],
        [('COOC', {0: 0, 1: 1, 2: 2, 3: 3})]
    ),
])
def test_identify_ptms(known_ptm_graphs, atoms, edges, expected):
    """
    Make sure PTMs are identified correctly.
    """
    molecule = make_molecule(atoms, edges)

    ptms = canmod.find_ptm_atoms(molecule)
    known_ptms = [(ptm_graph, nx.isomorphism.GraphMatcher(molecule, ptm_graph, node_match=canmod.ptm_node_matcher))
                  for ptm_graph in known_ptm_graphs]

    found = canmod.identify_ptms(molecule, ptms, known_ptms)
    found = [(ptm.name, match) for ptm, match in found]
    assert found == expected
